from textractor.entities.layout import Layout
from textractor.visualizers.entitylist import EntityList


class PageLayout:
    """
    Object representation of the layout components detected in the table.
    """

    def __init__(
        self,
        titles: EntityList[Layout] = EntityList([]),
        headers: EntityList[Layout] = EntityList([]),
        footers: EntityList[Layout] = EntityList([]),
        section_headers: EntityList[Layout] = EntityList([]),
        page_numbers: EntityList[Layout] = EntityList([]),
        lists: EntityList[Layout] = EntityList([]),
        figures: EntityList[Layout] = EntityList([]),
        tables: EntityList[Layout] = EntityList([]),
        key_values: EntityList[Layout] = EntityList([]),
    ):
        self._titles = titles
        self._headers = headers
        self._footers = footers
        self._section_headers = section_headers
        self._page_numbers = page_numbers
        self._lists = lists
        self._figures = figures
        self._tables = tables
        self._key_values = key_values

    @property
    def titles(self) -> EntityList[Layout]:
        """Titles detected in the Page

        :return: EntityList of titles detected in the page
        :rtype: EntityList[Layout]
        """
        return self._titles

    @property
    def headers(self) -> EntityList[Layout]:
        """Headers detected in the Page

        :return: EntityList of headers detected in the page
        :rtype: EntityList[Layout]
        """
        return self._headers

    @property
    def footers(self) -> EntityList[Layout]:
        """Footers detected in the Page

        :return: EntityList of footers detected in the page
        :rtype: EntityList[Layout]
        """
        return self._footers

    @property
    def section_headers(self) -> EntityList[Layout]:
        """Section headers detected in the Page

        :return: EntityList of section headers detected in the page
        :rtype: EntityList[Layout]
        """
        return self._section_headers

    @property
    def page_numbers(self) -> EntityList[Layout]:
        """Page numbers detected in the Page

        :return: EntityList of page numbers detected in the page
        :rtype: EntityList[Layout]
        """
        return self._page_numbers

    @property
    def lists(self) -> EntityList[Layout]:
        """Lists detected in the Page

        :return: EntityList of lists detected in the page
        :rtype: EntityList[Layout]
        """
        return self._lists

    @property
    def figures(self) -> EntityList[Layout]:
        """Figures detected in the Page

        :return: EntityList of figures detected in the page
        :rtype: EntityList[Layout]
        """
        return self._figures

    @property
    def tables(self) -> EntityList[Layout]:
        """Tables detected in the Page. This includes Tables detected by the AnalyzeDocument Tables API if used.

        :return: EntityList of tables detected in the page
        :rtype: EntityList[Layout]
        """
        return self._tables

    @property
    def key_values(self) -> EntityList[Layout]:
        """KeyValues detected in the Page

        :return: EntityList of keyvalues detected in the page
        :rtype: EntityList[Layout]
        """
        return self._key_values



"""
Represents a single :class:`Layout` Entity within the :class:`Document`.
The Textract API response returns groups of layout as LAYOUT_* BlockTypes.
"""

import os
from typing import List, Tuple
import uuid
from textractor.entities.bbox import BoundingBox
from textractor.entities.document_entity import DocumentEntity
from textractor.entities.line import Line
from textractor.entities.word import Word
from textractor.data.constants import (
    LAYOUT_SECTION_HEADER,
    LAYOUT_TITLE,
    LAYOUT_TABLE,
    LAYOUT_FIGURE,
    LAYOUT_FOOTER,
    LAYOUT_HEADER,
    LAYOUT_KEY_VALUE,
    LAYOUT_LIST,
    LAYOUT_PAGE_NUMBER,
    LAYOUT_TEXT,
    LAYOUT_ENTITY,
)
from textractor.data.text_linearization_config import TextLinearizationConfig
from textractor.utils.text_utils import group_elements_horizontally, linearize_children


class Layout(DocumentEntity):
    """
    To create a new :class:`Layout` object we need the following:
    """

    def __init__(
        self,
        entity_id: str,
        bbox: BoundingBox,
        reading_order: int,
        label: str,
        confidence: float = 0,
    ):
        super().__init__(entity_id, bbox)
        self.reading_order = reading_order
        self._confidence = confidence / 100
        self.layout_type = label
        self._page = None
        self._page_id = None

    @property
    def page(self):
        """
        :return: Returns the page number of the page the :class:`Layout` entity is present in.
        :rtype: int
        """
        return self._page

    @page.setter
    def page(self, page_num: int):
        """
        Sets the page number attribute of the :class:`Layout` entity.

        :param page_num: Page number where the :class:`Layout` entity exists.
        :type page_num: int
        """
        self._page = page_num

    @property
    def page_id(self) -> str:
        """
        :return: Returns the Page ID attribute of the page which the entity belongs to.
        :rtype: str
        """
        return self._page_id

    @page_id.setter
    def page_id(self, page_id: str):
        """
        Sets the Page ID of the :class:`Layout` entity.

        :param page_id: Page ID of the page the entity belongs to.
        :type page_id: str
        """
        self._page_id = page_id

    @property
    def text(self):
        return self.get_text()

    @property
    def words(self, config: TextLinearizationConfig = TextLinearizationConfig()):
        _, words = self.get_text_and_words(config)
        return words

    def get_text_and_words(
        self, config: TextLinearizationConfig = TextLinearizationConfig()
    ) -> Tuple[str, List[Word]]:
        """
        Returns the layout object text and words sorted in reading order

        :param config: Text linearization configuration object, defaults to TextLinearizationConfig()
        :type config: TextLinearizationConfig, optional
        :return: Tuple of page text and words
        :rtype: Tuple[str, List[Word]]
        """

        if (
            (self.layout_type == LAYOUT_HEADER and config.hide_header_layout)
            or (self.layout_type == LAYOUT_FOOTER and config.hide_footer_layout)
            or (self.layout_type == LAYOUT_FIGURE and config.hide_figure_layout)
            or (self.layout_type == LAYOUT_PAGE_NUMBER and config.hide_page_num_layout)
            or (self.layout_type == LAYOUT_TABLE and config.hide_table_layout)
            or (self.layout_type == LAYOUT_KEY_VALUE and config.hide_key_value_layout)
        ):
            return "", []
        elif self.layout_type == LAYOUT_PAGE_NUMBER:
            final_text, final_words = linearize_children(
                self.children,
                config,
                no_new_lines=False,
                is_layout_table=False,
            )
            if config.add_prefixes_and_suffixes_as_words:
                return (
                    config.page_num_prefix + final_text + config.page_num_suffix,
                    (
                        ([Word(str(uuid.uuid4()), BoundingBox.enclosing_bbox(final_words), is_structure=True), config.page_num_prefix] if config.page_num_prefix else []) +
                        final_words +
                        ([Word(str(uuid.uuid4()), BoundingBox.enclosing_bbox(final_words), is_structure=True), config.page_num_suffix] if config.page_num_suffix else [])
                    ),
                )
            else:
                return (
                    config.page_num_prefix + final_text + config.page_num_suffix,
                    final_words,
                )
        elif self.layout_type == LAYOUT_LIST:
            final_text = config.list_layout_prefix
            final_words = []
            for i, child in enumerate(
                sorted(self.children, key=lambda x: x.reading_order)
            ):
                child_text, child_words = child.get_text_and_words(config)
                final_text += (
                    (
                        config.list_element_prefix
                        if (
                            child_text[: len(config.list_element_prefix)] != config.list_element_prefix and
                            config.add_prefixes_and_suffixes_in_text
                        ) else ""
                    )
                    + (
                        child_text.replace("\n", " ")
                        if config.remove_new_lines_in_leaf_elements 
                        else child_text
                    )
                    + (
                        config.list_element_separator
                        if i != len(self.children) - 1
                        else ""
                    )
                )
                if config.add_prefixes_and_suffixes_as_words:
                    final_words += (
                        ([Word(str(uuid.uuid4(), BoundingBox.enclosing_bbox(child_words)), config.list_element_prefix, is_structure=True)] if config.list_element_prefix else []) +
                        child_words + 
                        ([Word(str(uuid.uuid4(), BoundingBox.enclosing_bbox(child_words)), config.list_element_suffix, is_structure=True)] if config.list_element_suffix else [])
                    )
                else:
                    final_words += child_words
            final_text += config.list_layout_suffix
        elif self.layout_type == LAYOUT_TITLE:
            final_text, final_words = linearize_children(
                self.children, config, no_new_lines=True
            )
            if config.add_prefixes_and_suffixes_in_text:
                final_text = config.title_prefix + final_text + config.title_suffix
            if config.add_prefixes_and_suffixes_as_words:
                final_words = (
                    ([Word(str(uuid.uuid4()), BoundingBox.enclosing_bbox(final_words), config.title_prefix, is_structure=True)] if config.title_prefix else []) + 
                    final_words + 
                    ([Word(str(uuid.uuid4()), BoundingBox.enclosing_bbox(final_words), config.title_suffix, is_structure=True)] if config.title_suffix else []) 
                )
        elif self.layout_type == LAYOUT_HEADER:
            final_text, final_words = linearize_children(
                self.children, config, no_new_lines=True
            )
            if config.add_prefixes_and_suffixes_in_text:
                final_text = (
                    config.header_prefix + final_text + config.header_suffix
                )
            if config.add_prefixes_and_suffixes_as_words:
                final_words = (
                    ([Word(str(uuid.uuid4()), BoundingBox.enclosing_bbox(final_words), config.header_prefix, is_structure=True)] if config.header_prefix else []) + 
                    final_words + 
                    ([Word(str(uuid.uuid4()), BoundingBox.enclosing_bbox(final_words), config.header_suffix, is_structure=True)] if config.header_suffix else []) 
                )
        elif self.layout_type == LAYOUT_SECTION_HEADER:
            final_text, final_words = linearize_children(
                self.children, config, no_new_lines=True
            )
            if config.add_prefixes_and_suffixes_in_text:
                final_text = (
                    config.section_header_prefix + final_text + config.section_header_suffix
                )
            if config.add_prefixes_and_suffixes_as_words:
                final_words = (
                    ([Word(str(uuid.uuid4()), BoundingBox.enclosing_bbox(final_words), config.section_header_prefix, is_structure=True)] if config.section_header_prefix else []) + 
                    final_words + 
                    ([Word(str(uuid.uuid4()), BoundingBox.enclosing_bbox(final_words), config.section_header_suffix, is_structure=True)] if config.section_header_suffix else []) 
                )
        elif self.layout_type == LAYOUT_TEXT:
            final_text, final_words = linearize_children(
                self.children,
                config,
                no_new_lines=True,
            )
            final_text = (
                config.text_prefix + final_text + config.text_suffix
            )
        else:
            final_text, final_words = linearize_children(
                self.children,
                config,
                no_new_lines=False,
                is_layout_table=self.layout_type == LAYOUT_TABLE,
            )

            if config.add_prefixes_and_suffixes_in_text:
                if self.layout_type == LAYOUT_TABLE:
                    final_text = (
                        config.table_layout_prefix + final_text + config.table_layout_suffix
                    )
                elif self.layout_type == LAYOUT_KEY_VALUE:
                    final_text = (
                        config.key_value_layout_prefix + final_text + config.key_value_layout_suffix
                    )
                elif self.layout_type == LAYOUT_FIGURE:
                    final_text = (
                        config.figure_layout_prefix + final_text + config.figure_layout_suffix
                    )
                elif self.layout_type == LAYOUT_ENTITY:
                    final_text = (
                        config.entity_layout_prefix + final_text + config.entity_layout_suffix
                    )
                elif self.layout_type == LAYOUT_FOOTER:
                    final_text = (
                        config.footer_layout_prefix + final_text + config.footer_layout_suffix
                    )
            if config.add_prefixes_and_suffixes_as_words:
                if self.layout_type == LAYOUT_TABLE:
                    final_words = (
                        ([Word(str(uuid.uuid4()), BoundingBox.enclosing_bbox(final_words), config.table_layout_prefix, is_structure=True)] if config.table_layout_prefix else []) + 
                        final_words + 
                        ([Word(str(uuid.uuid4()), BoundingBox.enclosing_bbox(final_words), config.table_layout_suffix, is_structure=True)] if config.table_layout_suffix else []) 
                    )
                elif self.layout_type == LAYOUT_KEY_VALUE:
                    final_words = (
                        ([Word(str(uuid.uuid4()), BoundingBox.enclosing_bbox(final_words), config.key_value_layout_prefix, is_structure=True)] if config.key_value_layout_prefix else []) + 
                        final_words + 
                        ([Word(str(uuid.uuid4()), BoundingBox.enclosing_bbox(final_words), config.key_value_layout_suffix, is_structure=True)] if config.key_value_layout_suffix else []) 
                    )
                elif self.layout_type == LAYOUT_FIGURE:
                    final_words = (
                        ([Word(str(uuid.uuid4()), BoundingBox.enclosing_bbox(final_words), config.figure_layout_prefix, is_structure=True)] if config.figure_layout_prefix else []) + 
                        final_words + 
                        ([Word(str(uuid.uuid4()), BoundingBox.enclosing_bbox(final_words), config.figure_layout_suffix, is_structure=True)] if config.figure_layout_suffix else []) 
                    )

        while (
            config.layout_element_separator * (config.max_number_of_consecutive_new_lines + 1) in final_text
        ):
            final_text = final_text.replace(
                "\n" * (config.max_number_of_consecutive_new_lines + 1),
                "\n" * config.max_number_of_consecutive_new_lines,
            )

        while config.max_number_of_consecutive_spaces and (
            " " * (config.max_number_of_consecutive_spaces + 1) in final_text
        ):
            final_text = final_text.replace(
                " " * (config.max_number_of_consecutive_spaces + 1),
                " " * config.max_number_of_consecutive_spaces,
            )

        return final_text, final_words



"""The Document class is defined to host all the various DocumentEntity objects within it. :class:`DocumentEntity` objects can be 
accessed, searched and exported the functions given below."""

import boto3
import json
import os
import string
import logging
import xlsxwriter
import io
from typing import List, IO, Union, AnyStr, Tuple
from copy import deepcopy
from collections import defaultdict
from PIL import Image

from textractor.entities.expense_document import ExpenseDocument
from textractor.entities.identity_document import IdentityDocument
from textractor.entities.word import Word
from textractor.entities.line import Line
from textractor.entities.page import Page
from textractor.entities.table import Table
from textractor.entities.query import Query
from textractor.entities.signature import Signature
from textractor.entities.layout import Layout
from textractor.exceptions import InputError
from textractor.entities.key_value import KeyValue
from textractor.entities.bbox import SpatialObject
from textractor.utils.s3_utils import download_from_s3
from textractor.visualizers.entitylist import EntityList
from textractor.data.constants import (
    TextTypes,
    SimilarityMetric,
    Direction,
    DirectionalFinderType,
)
from textractor.utils.search_utils import SearchUtils
from textractor.data.text_linearization_config import TextLinearizationConfig
from textractor.data.html_linearization_config import HTMLLinearizationConfig
from textractor.entities.linearizable import Linearizable


class Document(SpatialObject, Linearizable):
    """
    Represents the description of a single document, as it would appear in the input to the Textract API.
    Document serves as the root node of the object model hierarchy,
    which should be used as an intermediate form for most analytic purposes.
    The Document node also contains the metadata of the document.
    """

    @classmethod
    def open(cls, fp: Union[dict, str, IO[AnyStr]]):
        """Create a Document object from a JSON file path, file handle or response dictionary

        :param fp: _description_
        :type fp: Union[dict, str, IO[AnyStr]]
        :raises InputError: Raised on input not being of type Union[dict, str, IO[AnyStr]]
        :return: Document object
        :rtype: Document
        """
        from textractor.parsers import response_parser

        if isinstance(fp, dict):
            return response_parser.parse(fp)
        elif isinstance(fp, str):
            if fp.startswith("s3://"):
                # FIXME: Opening s3 clients for everythign should be avoided
                client = boto3.client("s3")
                return response_parser.parse(json.load(download_from_s3(client, fp)))
            with open(fp, "r") as f:
                return response_parser.parse(json.load(f))
        elif isinstance(fp, io.IOBase):
            return response_parser.parse(json.load(fp))
        else:
            raise InputError(
                f"Document.open() input must be of type dict, str or file handle, not {type(fp)}"
            )

    def __init__(self, num_pages: int = 1):
        """
        Creates a new document, ideally containing entity objects pertaining to each page.

        :param num_pages: Number of pages in the input Document.
        """
        super().__init__(width=0, height=0)
        self.num_pages: int = num_pages
        self._pages: List[Page] = []
        self._identity_documents: List[IdentityDocument] = []
        self._trp2_document = None
        self.response = None

    @property
    def words(self) -> EntityList[Word]:
        """
        Returns all the :class:`Word` objects present in the Document.

        :return: List of Word objects, each representing a word within the Document.
        :rtype: EntityList[Word]
        """
        return EntityList(sum([page.words for page in self.pages], []))

    @property
    def text(self) -> str:
        """Returns the document text as one string

        :return: Page text seperated by line return
        :rtype: str
        """
        return os.linesep.join([page.text for page in self.pages])

    @property
    def identity_documents(self) -> EntityList[IdentityDocument]:
        """
        Returns all the :class:`IdentityDocument` objects present in the Document.

        :return: List of IdentityDocument objects, each representing an identity document within the Document.
        :rtype: EntityList[IdentityDocument]
        """
        return EntityList(self._identity_documents)

    @identity_documents.setter
    def identity_documents(self, identity_documents: List[IdentityDocument]):
        """
        Set all the identity documents detected inside the Document
        """
        self._identity_documents = identity_documents

    @property
    def expense_documents(self) -> EntityList[ExpenseDocument]:
        """
        Returns all the :class:`ExpenseDocument` objects present in the Document.

        :return: List of ExpenseDocument objects, each representing an expense document within the Document.
        :rtype: EntityList[ExpenseDocument]
        """
        return EntityList(sum([page.expense_documents for page in self.pages], []))

    @property
    def lines(self) -> EntityList[Line]:
        """
        Returns all the :class:`Line` objects present in the Document.

        :return: List of Line objects, each representing a line within the Document.
        :rtype: EntityList[Line]
        """
        return EntityList(sum([page.lines for page in self.pages], []))

    @property
    def key_values(self) -> EntityList[KeyValue]:
        """
        Returns all the :class:`KeyValue` objects present in the Document.

        :return: List of KeyValue objects, each representing a key-value pair within the Document.
        :rtype: EntityList[KeyValue]
        """
        return EntityList(sum([page.key_values for page in self.pages], []))

    @property
    def checkboxes(self) -> EntityList[KeyValue]:
        """
        Returns all the :class:`KeyValue` objects with SelectionElements present in the Document.

        :return: List of KeyValue objects, each representing a checkbox within the Document.
        :rtype: EntityList[KeyValue]
        """
        return EntityList(sum([page.checkboxes for page in self.pages], []))

    @property
    def tables(self) -> EntityList[Table]:
        """
        Returns all the :class:`Table` objects present in the Document.

        :return: List of Table objects, each representing a table within the Document.
        :rtype: EntityList[Table]
        """
        return EntityList(sum([page.tables for page in self.pages], []))

    @property
    def queries(self) -> EntityList[Query]:
        """
        Returns all the :class:`Query` objects present in the Document.

        :return: List of Query objects.
        :rtype: EntityList[Query]
        """
        return EntityList(sum([page.queries for page in self.pages], []))

    @property
    def signatures(self) -> EntityList[Signature]:
        """
        Returns all the :class:`Signature` objects present in the Document.

        :return: List of Signature objects.
        :rtype: EntityList[Signature]
        """
        return EntityList(sum([page.signatures for page in self.pages], []))

    @property
    def layouts(self) -> EntityList[Layout]:
        """
        Returns all the :class:`Layout` objects present in the Document

        :return: List of Layout objects
        :rtype: EntityList[Layout]
        """
        return EntityList(sum([page.layouts for page in self.pages], []))

    @property
    def identity_document(self) -> EntityList[IdentityDocument]:
        """
        Returns all the :class:`IdentityDocument` objects present in the Page.

        :return: List of IdentityDocument objects.
        :rtype: EntityList
        """
        return EntityList(self._identity_documents)

    @identity_document.setter
    def identity_document(self, identity_documents: List[IdentityDocument]):
        """
        Add IdentityDocument objects to the Page.

        :param tables: List of IdentityDocument objects.
        :type identity_documents: list
        """
        self._identity_document = identity_documents

    @property
    def images(self) -> List[Image.Image]:
        """
        Returns all the page images in the Document.

        :return: List of PIL Image objects.
        :rtype: PIL.Image
        """
        return [page.image for page in self._pages]

    @property
    def pages(self) -> List[Page]:
        """
        Returns all the :class:`Page` objects present in the Document.

        :return: List of Page objects, each representing a Page within the Document.
        :rtype: List
        """
        return self._pages

    @pages.setter
    def pages(self, pages: List[Page]):
        """
        Add Page objects to the Document.

        :param pages: List of Page objects, each representing a page within the document.
        No specific ordering is assumed with input.
        :type pages: List[Page]
        """
        self._pages = sorted(pages, key=lambda x: x.page_num)

    def get_text_and_words(
        self, config: TextLinearizationConfig = TextLinearizationConfig()
    ) -> Tuple[str, List]:
        text, words_lists = zip(*[p.get_text_and_words(config) for p in self.pages])
        flattened_words = []
        for words in words_lists:
            flattened_words.extend(words)
        return config.layout_element_separator.join(text), flattened_words

    def page(self, page_no: int = 0):
        """
        Returns :class:`Page` object/s depending on the input page_no. Follows zero-indexing.

        :param page_no: if int, returns single Page Object, else if list, it returns a list of
                        Page objects.
        :type page_no: int if single page, list of int if multiple pages

        :return: Filters and returns Page objects depending on the input page_no
        :rtype: Page or List[Page]
        """
        if isinstance(page_no, int):
            return self.pages[page_no]
        elif isinstance(page_no, list):
            return [self.pages[num] for num in page_no]
        else:
            raise InputError("page_no parameter doesn't match required data type.")

    def to_html(self, config: HTMLLinearizationConfig = HTMLLinearizationConfig()):
        """
        Returns the HTML representation of the document, effectively calls Linearizable.to_html()
        but add <html><body></body></html> around the result and put each page in a <div>. 

        :return: HTML text of the entity
        :rtype: str
        """
        
        html = "<html><body>"
        for page in self.pages:
            html += f"<div>{page.to_html(config=config)}</div>"
        html += "</body></html>"
        
        return html

    def __repr__(self):
        return os.linesep.join(
            [
                "This document holds the following data:",
                f"Pages - {len(self.pages)}",
                f"Words - {len(self.words)}",
                f"Lines - {len(self.lines)}",
                f"Key-values - {len(self.key_values)}",
                f"Checkboxes - {len(self.checkboxes)}",
                f"Tables - {len(self.tables)}",
                f"Queries - {len(self.queries)}",
                f"Signatures - {len(self.signatures)}",
                f"Identity Documents - {len(self.identity_documents)}",
                f"Expense Documents - {len(self.expense_documents)}",
            ]
        )

    def to_trp2(self):
        """
        Parses the response to the trp2 format for backward compatibility

        :return: TDocument object that can be used with the older Textractor libraries
        :rtype: TDocument
        """
        from trp.trp2 import TDocument, TDocumentSchema
        
        if not self._trp2_document:
            self._trp2_document = TDocumentSchema().load(self.response)
        return self._trp2_document

    def visualize(self, *args, **kwargs):
        """
        Returns the object's children in a visualization EntityList object

        :return: Returns an EntityList object
        :rtype: EntityList
        """
        return EntityList(self.pages).visualize(*args, **kwargs)

    def keys(self, include_checkboxes: bool = True) -> List[str]:
        """
        Prints all keys for key-value pairs and checkboxes if the document contains them.

        :param include_checkboxes: True/False. Set False if checkboxes need to be excluded.
        :type include_checkboxes: bool

        :return: List of strings containing key names in the Document
        :rtype: List[str]
        """
        keys = []
        keys = [keyvalue.key for keyvalue in self.key_values]
        if include_checkboxes:
            keys += [keyvalue.key for keyvalue in self.checkboxes]
        return keys

    def filter_checkboxes(
        self, selected: bool = True, not_selected: bool = True
    ) -> List[KeyValue]:
        """
        Return a list of :class:`KeyValue` objects containing checkboxes if the document contains them.

        :param selected: True/False Return SELECTED checkboxes
        :type selected: bool
        :param not_selected: True/False Return NOT_SELECTED checkboxes
        :type not_selected: bool

        :return: Returns checkboxes that match the conditions set by the flags.
        :rtype: EntityList[KeyValue]
        """

        checkboxes = EntityList([])
        for page in self.pages:
            checkboxes.extend(
                page.filter_checkboxes(selected=selected, not_selected=not_selected)
            )
        return checkboxes

    def get_words_by_type(self, text_type: TextTypes = TextTypes.PRINTED) -> List[Word]:
        """
        Returns list of :class:`Word` entities that match the input text type.

        :param text_type: TextTypes.PRINTED or TextTypes.HANDWRITING
        :type text_type: TextTypes
        :return: Returns list of Word entities that match the input text type.
        :rtype: EntityList[Word]
        """
        if not self.words:
            logging.warn("Document contains no word entities.")
            return []

        filtered_words = EntityList()
        for page in self.pages:
            filtered_words.extend(page.get_words_by_type(text_type=text_type))
        return filtered_words

    def search_words(
        self,
        keyword: str,
        top_k: int = 1,
        similarity_metric: SimilarityMetric = SimilarityMetric.LEVENSHTEIN,
        similarity_threshold: float = 0.6,
    ) -> List[Word]:
        """
        Return a list of top_k words that match the keyword.

        :param keyword: Keyword that is used to query the document.
        :type keyword: str
        :param top_k: Number of closest word objects to be returned
        :type top_k: int
        :param similarity_metric: SimilarityMetric.COSINE, SimilarityMetric.EUCLIDEAN or SimilarityMetric.LEVENSHTEIN. SimilarityMetric.COSINE is chosen as default.
        :type similarity_metric: SimilarityMetric
        :param similarity_threshold: Measure of how similar document key is to queried key. default=0.6
        :type similarity_threshold: float

        :return: Returns a list of words that match the queried key sorted from highest
                 to lowest similarity.
        :rtype: EntityList[Word]
        """

        top_n_words = []
        for page in self.pages:
            top_n_words.extend(
                page._search_words_with_similarity(
                    keyword=keyword,
                    top_k=top_k,
                    similarity_metric=similarity_metric,
                    similarity_threshold=similarity_threshold,
                )
            )

        top_n_words = sorted(top_n_words, key=lambda x: x[0], reverse=True)[:top_k]
        top_n_words = EntityList([ent[1] for ent in top_n_words])

        return top_n_words

    def search_lines(
        self,
        keyword: str,
        top_k: int = 1,
        similarity_metric: SimilarityMetric = SimilarityMetric.LEVENSHTEIN,
        similarity_threshold: float = 0.6,
    ) -> List[Line]:
        """
        Return a list of top_k lines that contain the queried keyword.

        :param keyword: Keyword that is used to query the document.
        :type keyword: str
        :param top_k: Number of closest line objects to be returned
        :type top_k: int
        :param similarity_metric: SimilarityMetric.COSINE, SimilarityMetric.EUCLIDEAN or SimilarityMetric.LEVENSHTEIN. SimilarityMetric.COSINE is chosen as default.
        :type similarity_metric: SimilarityMetric
        :param similarity_threshold: Measure of how similar document key is to queried key. default=0.6
        :type similarity_threshold: float

        :return: Returns a list of lines that contain the queried key sorted from highest
                 to lowest similarity.
        :rtype: EntityList[Line]
        """
        if not isinstance(similarity_metric, SimilarityMetric):
            raise InputError(
                "similarity_metric parameter should be of SimilarityMetric type. Find input choices from textractor.data.constants"
            )

        top_n_lines = []
        for page in self.pages:
            top_n_lines.extend(
                page._search_lines_with_similarity(
                    keyword=keyword,
                    top_k=top_k,
                    similarity_metric=similarity_metric,
                    similarity_threshold=similarity_threshold,
                )
            )

        top_n_lines = EntityList([ent[1] for ent in top_n_lines][:top_k])

        return top_n_lines

    # KeyValue entity related functions
    def get(
        self,
        key: str,
        top_k_matches: int = 1,
        similarity_metric: SimilarityMetric = SimilarityMetric.LEVENSHTEIN,
        similarity_threshold: float = 0.6,
    ):
        """
        Return upto top_k_matches of key-value pairs for the key that is queried from the document.

        :param key: Query key to match
        :type key: str
        :param top_k_matches: Maximum number of matches to return
        :type top_k_matches: int
        :param similarity_metric: SimilarityMetric.COSINE, SimilarityMetric.EUCLIDEAN or SimilarityMetric.LEVENSHTEIN. SimilarityMetric.COSINE is chosen as default.
        :type similarity_metric: SimilarityMetric
        :param similarity_threshold: Measure of how similar document key is to queried key. default=0.6
        :type similarity_threshold: float

        :return: Returns a list of key-value pairs that match the queried key sorted from highest to lowest similarity.
        :rtype: EntityList[KeyValue]
        """
        if not isinstance(similarity_metric, SimilarityMetric):
            raise InputError(
                "similarity_metric parameter should be of SimilarityMetric type. Find input choices from textractor.data.constants"
            )

        top_n = []
        similarity_threshold = (
            -similarity_threshold
            if similarity_metric == SimilarityMetric.EUCLIDEAN
            else similarity_threshold
        )
        lowest_similarity = similarity_threshold

        for kv in self.key_values + self.checkboxes:
            try:
                edited_document_key = "".join(
                    [
                        char
                        for char in kv.key.__repr__()
                        if char not in string.punctuation
                    ]
                )
            except:
                pass
            key = "".join([char for char in key if char not in string.punctuation])

            similarity = [
                SearchUtils.get_word_similarity(key, word, similarity_metric)
                for word in edited_document_key.split(" ")
            ]
            similarity.append(
                SearchUtils.get_word_similarity(
                    key, edited_document_key, similarity_metric
                )
            )

            similarity = (
                min(similarity)
                if similarity_metric == SimilarityMetric.EUCLIDEAN
                else max(similarity)
            )

            if similarity > similarity_threshold:
                if len(top_n) < top_k_matches:
                    top_n.append((kv, similarity))
                elif similarity > lowest_similarity:
                    top_n[-1] = (kv, similarity)
                top_n = sorted(top_n, key=lambda x: x[1], reverse=True)
                lowest_similarity = top_n[-1][1]

        if not top_n:
            logging.warning(
                f"Query key does not match any existing keys in the document.{os.linesep}{self.keys()}"
            )
            return EntityList([])

        logging.info(f"Query key matched {len(top_n)} key-values in the document.")

        return EntityList([value[0] for value in top_n])

    # Export document entities into supported formats
    def export_kv_to_csv(
        self,
        include_kv: bool = True,
        include_checkboxes: bool = True,
        filepath: str = "Key-Values.csv",
    ):
        """
        Export key-value entities and checkboxes in csv format.

        :param include_kv: True if KVs are to be exported. Else False.
        :type include_kv: bool
        :param include_checkboxes: True if checkboxes are to be exported. Else False.
        :type include_checkboxes: bool
        :param filepath: Path to where file is to be stored.
        :type filepath: str
        """
        keys = []
        values = []
        if include_kv and not self.key_values:
            logging.warning("Document does not contain key-values.")
        elif include_kv:
            for kv in self.key_values:
                keys.append(" ".join([w.text for w in kv.key]))
                values.append(kv.value.get_text())

        if include_checkboxes and not self.checkboxes:
            logging.warning("Document does not contain checkbox elements.")
        elif include_checkboxes:
            for kv in self.checkboxes:
                keys.append(" ".join([w.text for w in kv.key]))
                values.append(kv.value.children[0].status.name)

        with open(filepath, "w") as f:
            f.write(f"Key;Value{os.linesep}")
            for k, v in zip(keys, values):
                f.write(f"{k};{v}{os.linesep}")

        logging.info(
            f"csv file stored at location {os.path.join(os.getcwd(),filepath)}"
        )

    def export_kv_to_txt(
        self,
        include_kv: bool = True,
        include_checkboxes: bool = True,
        filepath: str = "Key-Values.txt",
    ):
        """
        Export key-value entities and checkboxes in txt format.

        :param include_kv: True if KVs are to be exported. Else False.
        :type include_kv: bool
        :param include_checkboxes: True if checkboxes are to be exported. Else False.
        :type include_checkboxes: bool
        :param filepath: Path to where file is to be stored.
        :type filepath: str
        """
        export_str = ""
        index = 1
        if include_kv and not self.key_values:
            logging.warning("Document does not contain key-values.")
        elif include_kv:
            for kv in self.key_values:
                export_str += (
                    f"{index}. {kv.key.__repr__()} : {kv.value.__repr__()}{os.linesep}"
                )
                index += 1

        if include_checkboxes and not self.checkboxes:
            logging.warning("Document does not contain checkbox elements.")
        elif include_checkboxes:
            for kv in self.checkboxes:
                export_str += f"{index}. {kv.key.__repr__()} : {kv.value.children[0].status.name}{os.linesep}"
                index += 1

        with open(filepath, "w") as text_file:
            text_file.write(export_str)
        logging.info(
            f"txt file stored at location {os.path.join(os.getcwd(),filepath)}"
        )

    def export_tables_to_excel(self, filepath):
        """
        Creates an excel file and writes each table on a separate worksheet within the workbook.
        This is stored on the filepath passed by the user.

        :param filepath: Path to store the exported Excel file.
        :type filepath: str, required
        """
        if not filepath:
            logging.error("Filepath required to store excel file.")
        workbook = xlsxwriter.Workbook(filepath)
        for table in self.tables:
            workbook = table.to_excel(
                filepath=None, workbook=workbook, save_workbook=False
            )
        workbook.close()

    def independent_words(self):
        """
        :return: Return all words in the document, outside of tables, checkboxes, key-values.
        :rtype: EntityList[Word]
        """
        if not self.words:
            logging.warning("Words have not been assigned to this Document object.")
            return []

        else:
            table_words = sum([table.words for table in self.tables], [])
            kv_words = sum([kv.words for kv in self.key_values], [])
            checkbox_words = sum([kv.words for kv in self.checkboxes], [])
            dependent_words = table_words + checkbox_words + kv_words
            dependent_word_ids = set([word.id for word in dependent_words])
            independent_words = [
                word for word in self.words if word.id not in dependent_word_ids
            ]
            return EntityList(independent_words)

    def return_duplicates(self):
        """
        Returns a dictionary containing page numbers as keys and list of :class:`EntityList` objects as values.
        Each :class:`EntityList` instance contains the key-values and the last item is the table which contains duplicate information.
        This function is intended to let the Textract user know of duplicate objects extracted by the various Textract models.

        :return: Dictionary containing page numbers as keys and list of EntityList objects as values.
        :rtype: Dict[page_num, List[EntityList[DocumentEntity]]]
        """
        document_duplicates = defaultdict(list)

        for page in self.pages:
            document_duplicates[page.page_num].extend(page.return_duplicates())

        return document_duplicates

    def directional_finder(
        self,
        word_1: str = "",
        word_2: str = "",
        page: int = -1,
        prefix: str = "",
        direction=Direction.BELOW,
        entities=[],
    ):
        """
        The function returns entity types present in entities by prepending the prefix provided by te user. This helps in cases of repeating
        key-values and checkboxes. The user can manipulate original data or produce a copy. The main advantage of this function is to be able to define direction.

        :param word_1: The reference word from where x1, y1 coordinates are derived
        :type word_1: str, required
        :param word_2: The second word preferably in the direction indicated by the parameter direction. When it isn't given the end of page coordinates are used in the given direction.
        :type word_2: str, optional
        :param page: page number of the page in the document to search the entities in.
        :type page: int, required
        :param prefix: User provided prefix to prepend to the key . Without prefix, the method acts as a search by geometry function
        :type prefix: str, optional
        :param entities: List of DirectionalFinderType inputs.
        :type entities: List[DirectionalFinderType]

        :return: Returns the EntityList of modified key-value and/or checkboxes
        :rtype: EntityList
        """

        if not word_1 or page == -1:
            return EntityList([])

        x1, x2, y1, y2 = self._get_coords(word_1, word_2, direction, page)

        if x1 == -1:
            return EntityList([])

        page_obj = self.pages[page - 1]
        entity_dict = {
            DirectionalFinderType.KEY_VALUE_SET: self.key_values,
            DirectionalFinderType.SELECTION_ELEMENT: self.checkboxes,
        }

        entitylist = []
        for entity_type in entities:
            entitylist.extend(list(entity_dict[entity_type]))

        new_key_values = self._get_kv_with_direction(
            direction, entitylist, (x1, x2, y1, y2)
        )

        final_kv = []
        for kv in new_key_values:
            if kv.key:
                key_words = [deepcopy(word) for word in kv.key]
                key_words[0].text = prefix + key_words[0].text
                new_kv = deepcopy(kv)
                new_kv.key = key_words
                final_kv.append(new_kv)
            else:
                final_kv.append(kv)

        return EntityList(final_kv)

    def _get_kv_with_direction(self, direction, entitylist, coords):
        """Return key-values and checkboxes in entitylist present in the direction given with respect to the coordinates."""
        if direction == Direction.ABOVE:
            new_key_values = [
                kv
                for kv in entitylist
                if kv.bbox.y <= coords[2] and kv.bbox.y >= coords[-1]
            ]

        elif direction == Direction.BELOW:
            new_key_values = [
                kv
                for kv in entitylist
                if kv.bbox.y >= coords[2] and kv.bbox.y <= coords[-1]
            ]

        elif direction == Direction.RIGHT:
            new_key_values = [
                kv
                for kv in entitylist
                if kv.bbox.x >= coords[0] and kv.bbox.x <= coords[1]
            ]
            new_key_values = [
                kv
                for kv in new_key_values
                if kv.bbox.y >= coords[2] - kv.bbox.height
                and kv.bbox.y <= coords[-1] + 3 * kv.bbox.height
            ]

        elif direction == Direction.LEFT:
            new_key_values = [
                kv
                for kv in entitylist
                if kv.bbox.x <= coords[0] and kv.bbox.x >= coords[1]
            ]
            new_key_values = [
                kv
                for kv in new_key_values
                if kv.bbox.y >= coords[2] - kv.bbox.height
                and kv.bbox.y <= coords[-1] + 3 * kv.bbox.height
            ]

        return new_key_values

    def _get_coords(self, word_1, word_2, direction, page):
        """
        Returns coordinates for the area within which to search for key-values with the directional_finder by retrieving coordinates of word_1 \
        and word_2 if it exists else end of page.
        """
        word_1_objects = self.search_lines(
            keyword=word_1,
            top_k=5,
            similarity_metric=SimilarityMetric.COSINE,
            similarity_threshold=0.5,
        )
        word_1_objects = (
            [word for word in word_1_objects if word.page == page] if page != -1 else []
        )

        if not word_1_objects:
            logging.warning(f"{word_1} not found in page {page}")
            return -1, -1, -1, -1
        else:
            word_1_obj = word_1_objects[0]
            x1, y1 = word_1_obj.bbox.x, word_1_obj.bbox.y

        if word_2:
            word_2_objects = self.search_lines(
                keyword=word_2,
                top_k=5,
                similarity_metric=SimilarityMetric.COSINE,
                similarity_threshold=0.5,
            )
            word_2_objects = [word for word in word_2_objects if word.page == page]
            if not word_2_objects:
                logging.warning(f"{word_2} not found in page {page}")
                return -1, -1, -1, -1
            else:
                word_2_obj = word_2_objects[0]
                x2, y2 = word_2_obj.bbox.x, word_2_obj.bbox.y
        else:
            x2, y2 = x1, y1

        if direction == Direction.ABOVE:
            x1, x2, y1, y2 = (x1, x2, y1, y2) if y2 < y1 else (x1, 0, y1, 0)

        elif direction == Direction.BELOW:
            x1, x2, y1, y2 = (x1, x2, y1, y2) if y2 > y1 else (x1, 1, y1, 1)

        elif direction == Direction.RIGHT:
            x1, x2, y1, y2 = (x1, x2, y1, y2) if x2 > x1 else (x1, 1, y1, y1)

        elif direction == Direction.LEFT:
            x1, x2, y1, y2 = (x1, x2, y1, y2) if x2 < x1 else (x1, 0, y1, y1)

        else:
            return -1, -1, -1, -1

        return x1, x2, y1, y2



"""Utility functions for Document search"""

try:
    import numpy as np
except ImportError:
    # No need to log it here as numpy is only used if SentenceTransformers is used
    # The latter has numpy as dependency.
    pass

import math
import editdistance
from textractor.data.constants import SimilarityMetric
from textractor.exceptions import MissingDependencyException


from textractor.data.constants import (
    IS_COLUMN_HEAD,
    IS_FOOTER_CELL,
    IS_TITLE_CELL,
    IS_SUMMARY_CELL,
    IS_SECTION_TITLE_CELL,
    CellTypes,
)


class SearchUtils:
    model = None
    util = None
    model_string = "all-MiniLM-L6-v2"

    @classmethod
    def get_word_similarity(
        cls, word_1: str, word_2: str, similarity_metric: SimilarityMetric
    ) -> float:
        """
        Returns the extent of similarity between the input words using the similarity_metric input by the user.

        :param word_1: First word to check for similarity
        :type word_1: str
        :param word_2: Second word to check for similarity
        :type word_2: str
        :param similarity_metric: The function supports one of 3 metrics \
                                * Levenshtein distance/ edit distance \
                                * Euclidean distance \
                                * Cosine distance
        :type similarity_metric: str

        :return: Returns the similarity measure calculated based on the metric for the 2 input words.
        :rtype: float
        """
        if cls.model is None and similarity_metric != SimilarityMetric.LEVENSHTEIN:
            try:
                from sentence_transformers import SentenceTransformer, util
            except ImportError:
                raise MissingDependencyException(
                    "sentence_transformers is not installed. Use SimilarityMetric.LEVENSHTEIN."
                )
            cls.model = SentenceTransformer(cls.model_string)
            cls.util = util

        if similarity_metric == SimilarityMetric.LEVENSHTEIN:
            return normalized_edit_distance(word_1.lower(), word_2.lower())
        elif similarity_metric == SimilarityMetric.EUCLIDEAN:
            ref_word_emb = cls.model.encode([word_1])
            word_emb = cls.model.encode([word_2])
            dist = np.linalg.norm(ref_word_emb - word_emb)
            return dist
        else:
            ref_word_emb = cls.model.encode([word_1])
            word_emb = cls.model.encode([word_2])
            similarity = cls.util.cos_sim(ref_word_emb, word_emb)
            return similarity.item()


def jaccard_similarity(list_1: list, list_2: list) -> float:
    """
    Calculates Jaccard similarity between the 2 input lists.

    :param list_1: First list to check for similarity
    :type list_1: list
    :param list_2: Second list to check for similarity
    :type list_2: list

    :return: Returns the similarity measure calculated for the 2 input lists.
    :rtype: float
    """

    set_1 = set(list_1)
    set_2 = set(list_2)
    return float(len(set_1.intersection(set_2)) / len(set_1.union(set_2)))


def get_metadata_attr_name(cell_atr):
    """
    Returns metadata attribute mapping to the input CellType.

    :param cell_atr: Input cell type
    :type: enum
    :return: Returns metadata attribute mapping to the input CellType.
    :rtype: str
    """
    cell_map = {
        CellTypes.COLUMN_HEADER: IS_COLUMN_HEAD,
        CellTypes.SECTION_TITLE: IS_SECTION_TITLE_CELL,
        CellTypes.SUMMARY_CELL: IS_SUMMARY_CELL,
        CellTypes.FLOATING_TITLE: IS_TITLE_CELL,
        CellTypes.FLOATING_FOOTER: IS_FOOTER_CELL,
    }
    try:
        return cell_map[cell_atr]
    except:
        return ""


def normalized_edit_distance(s1: str, s2: str):
    """
    Returns the normalized edit distance

    :param s1: First string
    :type s1: str
    :param s2: Second string
    :type s2: str
    """

    dist = editdistance.eval(s1, s2)
    max_length = max(len(s1), len(s2))
    if max_length - dist == 0:
        return 0.0
    return (max_length - dist) / max_length



"""
The :class:`KeyValue` entity is a document entity representing the Forms output. The key in :class:`KeyValue` are typically words 
and the :class:`Value` could be :class:`Word` elements or :class:`SelectionElement` in case of checkboxes.

This class contains the associated metadata with the :class:`KeyValue` entity including the entity ID, 
bounding box information, value, existence of checkbox, page number, Page ID and confidence of detection.
"""

import logging
from typing import List
import uuid

from textractor.entities.word import Word
from textractor.entities.value import Value
from textractor.exceptions import InputError
from textractor.entities.bbox import BoundingBox
from textractor.entities.document_entity import DocumentEntity
from textractor.data.constants import TextTypes
from textractor.data.text_linearization_config import TextLinearizationConfig
from textractor.visualizers.entitylist import EntityList


class KeyValue(DocumentEntity):
    """
    To create a new :class:`KeyValue` object we require the following:

    :param entity_id: Unique identifier of the KeyValue entity.
    :type entity_id: str
    :param bbox: Bounding box of the KeyValue entity.
    :type bbox: BoundingBox
    :param contains_checkbox: True/False to indicate if the value is a checkbox.
    :type contains_checkbox: bool
    :param value: Value object that maps to the KeyValue entity.
    :type value: Value
    :param confidence: confidence with which the entity was detected.
    :type confidence: float
    """

    def __init__(
        self,
        entity_id: str,
        bbox: BoundingBox,
        contains_checkbox: bool = False,
        value: Value = None,
        confidence: float = 0,
    ):
        super().__init__(entity_id, bbox)

        self._words: EntityList[Word] = []
        self.contains_checkbox = contains_checkbox
        self._confidence = confidence / 100
        self._value: Value = value
        self.selection_status = False
        self._page = None
        self._page_id = None

    @property
    def ocr_confidence(self):
        """
        Return the average OCR confidence
        :return:
        """
        key_confidences = [w.confidence for w in self.key]
        value_confidences = (
            [w.confidence for w in self.value.words] if self.value else []
        )

        key_confidence = (
            sum(key_confidences) / len(key_confidences) if key_confidences else None
        )
        value_confidence = (
            sum(value_confidences) / len(value_confidences)
            if value_confidences
            else None
        )
        return {
            "key_average": key_confidence,
            "key_min": min(key_confidences) if key_confidences else None,
            "value_average": value_confidence,
            "value_min": min(value_confidences) if value_confidences else None,
        }

    @property
    def words(self) -> List[Word]:
        """
        Returns all the :class:`Word` objects present in the key and value of the :class:`KeyValue` object.

        :return words: List of Word objects, each representing a word within the KeyValue entity.
        :rtype: EntityList[Word]
        """
        value_words = self.value.words if self.value is not None and not self.contains_checkbox else []
        return EntityList(
            sorted(self._words + value_words, key=lambda x: x.bbox.x + x.bbox.y)
        )

    @property
    def key(self):
        """
        :return: Returns :class:`EntityList[Word]` object (a list of words) associated with the key.
        :rtype: EntityList[Word]
        """
        if not self._words:
            logging.info("Key contains no words objects.")
            return []
        return self._words

    @key.setter
    def key(self, words: List[Word]):
        """
        Add :class:`Word` objects to the Key.

        :param words: List of Word objects, each representing a word within the Key.
        No specific ordering is assumed.
        :type words: list
        """
        self._words = EntityList(words)

    @property
    def value(self) -> Value:
        """
        :return: Returns the :class:`Value` mapped to the key if it has been assigned.
        :rtype: Value
        """
        if self._value is None:
            logging.warning(
                "Asked for a value but it was never attributed "
                "-> make sure to assign value to key with the `kv.value = <Value Object>` property setter"
            )
            return None
        else:
            return self._value

    @value.setter
    def value(self, value: Value):
        """
        Add :class:`Value` object to the :class:`KeyValue`.

        :param value: Value object, representing a single value associated with the key.
        :type value: Value
        """
        self._value = value

    @property
    def page(self) -> int:
        """
        :return: Returns the page number of the page the :class:`Table` entity is present in.
        :rtype: int
        """
        return self._page

    @page.setter
    def page(self, page_num: int):
        """
        Sets the page number attribute of the :class:`Table` entity.

        :param page_num: Page number where the Table entity exists.
        :type page_num: int
        """
        self._page = page_num

    @property
    def page_id(self) -> str:
        """
        :return: Returns the Page ID attribute of the page which the entity belongs to.
        :rtype: str
        """
        return self._page_id

    @page_id.setter
    def page_id(self, page_id: str):
        """
        Sets the Page ID of the :class:`Table` entity.

        :param page_id: Page ID of the page the entity belongs to.
        :type page_id: str
        """
        self._page_id = page_id

    def get_words_by_type(self, text_type: str = TextTypes.PRINTED) -> List[Word]:
        """
        Returns list of :class:`Word` entities that match the input text type.

        :param text_type: TextTypes.PRINTED or TextTypes.HANDWRITING
        :type text_type: TextTypes

        :return: Returns list of Word entities that match the input text type.
        :rtype: EntityList[Word]
        """
        if not isinstance(text_type, TextTypes):
            raise InputError(
                "text_type parameter should be of TextTypes type. Find input choices from textractor.data.constants"
            )

        if not self.words:
            logging.info("Document contains no word entities.")
            return []
        else:
            return EntityList(
                [word for word in self.words if word.text_type == text_type]
            )

    def is_selected(self) -> bool:
        """
        For KeyValues containing a selection item, returns its `is_selected` status

        :return: Selection status of a selection item key value pair
        :rtype: bool
        """
        if self.contains_checkbox:
            if len(self.value.children) == 1:
                return self.value.children[0].is_selected()
            else:
                logging.info(
                    "is_checked() was called on a KeyValue that contains more than one checkbox. Returning first checkbox"
                )
                return self.value.children[0].is_selected()
        else:
            logging.info(
                "is_checked() was called on a KeyValue that does not contain checkboxes. Returning False"
            )
            return False

    def get_text_and_words(
        self, config: TextLinearizationConfig = TextLinearizationConfig()
    ):
        key_text, key_words = " ".join([w.text for w in self.key]), self.key
        value_text, value_words = self.value.get_text_and_words(config) if self.value else ("", "")
        words = []
        if not len(key_text) and not len(value_text):
            return "", []
        s = config.key_suffix.strip()
        key_suffix = (
            config.key_suffix
            if not (len(key_text) > len(s) and key_text[-len(s) :] == s)
            else " "
        )
        if config.add_prefixes_and_suffixes_in_text:
            text = f"{config.key_value_prefix}{config.key_prefix}{key_text}{key_suffix}{value_text}{config.key_value_suffix}"
        else:
            text = f"{key_text}{config.same_paragraph_separator}{value_text}"

        if config.add_prefixes_and_suffixes_as_words:
            words += [Word(str(uuid.uuid4()), self.bbox, config.key_value_prefix, is_structure=True)] if config.key_value_prefix else []
            if key_words:
                words += (
                    ([Word(str(uuid.uuid4()), BoundingBox.enclosing_bbox(self.key), config.key_prefix, is_structure=True)] if config.key_prefix else []) +
                    key_words +
                    ([Word(str(uuid.uuid4()), BoundingBox.enclosing_bbox(self.key), config.key_suffix, is_structure=True)] if config.key_suffix else [])
                )
            if value_words: 
                words += value_words
            words += ([Word(str(uuid.uuid4()), self.bbox, config.key_value_suffix, is_structure=True)] if config.key_value_suffix else [])
        else:
            words += key_words + value_words

        for w in words:
            w.kv_id = str(self.id)
            w.kv_bbox = self.bbox

        return text, words

    def __repr__(self) -> str:
        """
        :return: Returns KeyValue object as a formatted string.
        :rtype: str
        """
        repr_string = " ".join([word.text for word in self._words])

        if self.contains_checkbox:
            repr_string = self.value.__repr__() + " " + repr_string
        else:
            repr_string = repr_string
            repr_string += " : "
            repr_string += self.value.__repr__()
        return repr_string



from functools import cmp_to_key
import os
from typing import List, Tuple

from textractor.data.text_linearization_config import TextLinearizationConfig
from textractor.entities.bbox import BoundingBox
from textractor.entities.line import Line
from textractor.entities.word import Word
from textractor.entities.document_entity import DocumentEntity

def compare_bounding_box(a, b):
    ha = a.bbox.height
    hb = b.bbox.height

    delta = (ha + hb) / 3.5

    ay_mid = a.bbox.y + (ha / 2.0)
    by_mid = b.bbox.y + (hb / 2.0)

    if abs(ay_mid - by_mid) < delta:
        if a.bbox.x > b.bbox.x:
            return 1
        else:
            return -1

    else:
        if ay_mid > by_mid:
            return 1
        else:
            return -1

def group_elements_horizontally(
    elements: List[DocumentEntity], overlap_ratio: float = 0.5
) -> List[List[DocumentEntity]]:
    """
    Groups DocumentEntity objects based on their height using an overlap heuristic

    :param elements: DocumentEntity list
    :type elements: List[DocumentEntity]
    :param overlap_ratio: Height overlap ratio over which unaligned elements will be combined into a single line, defaults to 0.5
    :type overlap_ratio: float, optional
    :return: DocumentEntity sub-groups
    :rtype: List[List[DocumentEntity]]
    """
    sorted_elements = sorted(elements, key=cmp_to_key(compare_bounding_box))

    def vertical_overlap(line1, line2):
        top = max(line1.bbox.y, line2.bbox.y)
        bottom = min(line1.bbox.y + line1.bbox.height, line2.bbox.y + line2.bbox.height)
        return max(bottom - top, 0)

    def should_group(line, group):
        if not group:
            return False
        max_height = max(l.bbox.height for l in group)
        total_overlap = sum(vertical_overlap(line, l) for l in group)
        return total_overlap / max_height >= overlap_ratio

    grouped_elements = []
    if not sorted_elements:
        return grouped_elements

    current_group = [sorted_elements[0]]
    for element in sorted_elements[1:]:
        if "Table" in element.__class__.__name__:
            if current_group:
                grouped_elements.append(current_group)
            grouped_elements.append([element])
            current_group = []
        elif not current_group:  # if the group is emtpy add the line
            current_group.append(element)
        elif should_group(element, current_group):
            current_group.append(element)
        else:
            grouped_elements.append(current_group)
            current_group = [element]
    grouped_elements.append(current_group)
    return grouped_elements


def linearize_children(
    elements: List[DocumentEntity],
    config: TextLinearizationConfig = TextLinearizationConfig(),
    no_new_lines: bool = False,
    is_layout_table: bool = False,
) -> Tuple[str, List[Word]]:
    """
    Convert the words in this group into lines, keeping all words from the same lines together
    For example this is useful when trying to keep together words from the same cell that might be from lines
    Spanning across several cells

    :param elements: _description_
    :type elements: List[DocumentEntity]
    :param config: Text linearization configuration object, defaults to TextLinearizationConfig()
    :type config: TextLinearizationConfig, optional
    :param no_new_lines: Removes new lines in the output (single paragrph mode), defaults to False
    :type no_new_lines: bool, optional
    :param is_layout_table: If the element is a layout table we use the column separator instead of the regular separator, defaults to False
    :type is_layout_table: bool, optional
    :return: Tuple of text and linearized words
    :rtype: Tuple[str, List[Word]]
    """

    words_in_elements = [e for e in elements if isinstance(e, Word)]
    lines = set([w.line for w in words_in_elements])
    new_lines = []
    for line in lines:
        if line is None:
            continue
        new_lines.append(
            Line(
                line.id,
                line.bbox,
                sorted(
                    [w for w in words_in_elements if w.line is not None and w.line.id == line.id],
                    key=lambda x: x.bbox.x,
                ),
            )
        )
    elements = [e for e in elements if not isinstance(e, Word)] + new_lines
    grouped_elements = group_elements_horizontally(
        elements, config.heuristic_overlap_ratio
    )

    def part_of_same_paragraph(element1, element2, config=config):
        if (
            "Line" in element1.__class__.__name__
            and "Line" in element2.__class__.__name__
        ):
            return abs(
                element1.bbox.x - element2.bbox.x
            ) <= config.heuristic_h_tolerance * element1.bbox.width and abs(
                element1.bbox.y + element1.bbox.height - element2.bbox.y
            ) <= config.heuristic_line_break_threshold * min(
                element1.bbox.height, element2.bbox.height
            )
        return False

    result = ""
    words_output = []
    prev_element = None

    for group_idx, group in enumerate(grouped_elements):
        sorted_group = sorted(group, key=lambda element: element.bbox.x)
        if not sorted_group:
            continue

        added_words = set()
        for idx, element in enumerate(sorted_group):
            text_element, words_element = element.get_text_and_words(config)
            if "Table" in element.__class__.__name__ and len(words_element):
                result += text_element
                for w in words_element:
                    added_words.add(w.id)
                words_output += words_element
            elif "KeyValue" in element.__class__.__name__ and len(words_element):
                separator = (
                    config.same_paragraph_separator
                    if prev_element and part_of_same_paragraph(prev_element, element, config) else
                    config.same_layout_element_separator
                )
                result += separator + text_element
                for w in words_element:
                    added_words.add(w.id)
                words_output += words_element
            elif prev_element is None:
                result += text_element
                words_output += words_element
            elif is_layout_table:
                result += (
                    "" if idx == 0 else config.table_column_separator
                ) + text_element
                words_output += words_element
            elif part_of_same_paragraph(prev_element, element, config):
                result += config.same_paragraph_separator + text_element
                words_output += words_element
            else:
                result += config.same_layout_element_separator + text_element
                words_output += words_element

            # FIXME: Seems like this would be mostly needed
            # result += os.linesep
            prev_element = element

        if is_layout_table:
            result += config.table_row_separator
        else:
            result += config.same_layout_element_separator

        # We make a dummy line element with the bbox from the previous group
        prev_element = Line(
            "", BoundingBox.enclosing_bbox([o.bbox for o in sorted_group])
        )

    output = "".join(result)

    if no_new_lines:
        while os.linesep in output:
            output = output.replace(os.linesep, " ")
        while "  " in output:
            output = output.replace("  ", " ")

    return output, words_output


def is_distinct_entity(entity1: DocumentEntity, entity2: DocumentEntity) -> bool:
    """
    Check whether two DocumentEntity have word overlap

    :param entity1: First DocumentEntity
    :type entity1: DocumentEntity
    :param entity2: Second DocumentEntity
    :type entity2: DocumentEntity
    :return: Entities overlap as a boolean
    :rtype: bool
    """
    for w in entity1.words():
        if w in entity2.words():
            return False
    return True



import argparse
import json
import os

import yaml
from textractor.data.text_linearization_config import TextLinearizationConfig

from textractor import Textractor
from textractor.data.constants import (
    TextractAPI,
    TextractFeatures,
    CLIPrint,
    CLIOverlay,
)
from textractor.exceptions import InputError
from textractor.visualizers.entitylist import EntityList

STRING_DICT = {
    # Textractor parameters
    "PROFILE_NAME": "AWS profile name to use for the request",
    "REGION": "AWS region to use for the request",
    # Subcommand decriptions
    "DETECT_DOCUMENT_TEXT": "Synchronous API for Optical Character Recognition",
    "START_DOCUMENT_TEXT_DETECTION": "Asynchronous API for Optical Character Recognition",
    "ANALYZE_DOCUMENT": "Synchronous API for document analysis (forms, tables, queries, and signatures)",
    "START_DOCUMENT_ANALYSIS": "Asynchronous API for document analysis (forms, tables, queries, and signatures)",
    "ANALYZE_EXPENSE": "Synchronous API for expense analysis",
    "START_EXPENSE_ANALYSIS": "Asynchronous API for expense analysis",
    "ANALYZE_ID": "API for identity document analysis (supports driver's license and passports).",
    "GET_RESULT": "Try to fetch the result for a given job id",
    # Parameter descriptions
    "OUTPUT_FILE": "Output file to save the response, can be an S3 path",
    "INPUT_FILE_SYNC": "File to process, must be of type JPEG, PNG, TIFF, BMP. Can be an S3 path",
    "INPUT_FILE_ASYNC": "File to process, must be of type PDF, JPEG, PNG, TIFF, BMP. The file has to be in S3, you you can provide an S3 path with --upload-s3-path",
    "GET_RESULT_JOB_ID": "Job ID, as returned by any of the asynchronous functions",
    "QUERIES": 'List of queries, use quotes (") to escape spaces',
    "S3_UPLOAD_PATH": "Path to upload the input files to, required if input_file is not an S3 path",
    "S3_OUTPUT_PATH": "Path to write the response to",
    "GET_RESULT_API": "API used to make the request",
    # Output print
    "PRINT": "Print the output in a readable format",
    # Output linearize
    "LINEARIZE": "Print the linearized document output",
    "LINEARIZE_CONFIG": "Configuration file for the linearization",
    # Overlay
    "OVERLAY": "Save an image of the document with the words, lines, form fields, and tables overlayed on top",
    "FONT_SIZE_RATIO": "Scales the text up or down, default is 0.75, which would be half the pixel height",
}


def _build_parser():
    parser = argparse.ArgumentParser(
        description="Commandline interface for the Textractor library"
    )

    subparsers = parser.add_subparsers(help="Sub-command help", dest="subcommand")

    # DocumentText parsers
    parser_detect_document_text = subparsers.add_parser(
        "detect-document-text", help=STRING_DICT["DETECT_DOCUMENT_TEXT"]
    )
    parser_detect_document_text.add_argument(
        "file_source", type=str, help=STRING_DICT["INPUT_FILE_SYNC"]
    )
    parser_detect_document_text.add_argument(
        "output_file", type=str, help=STRING_DICT["OUTPUT_FILE"]
    )
    parser_detect_document_text.add_argument(
        "--profile-name", type=str, help=STRING_DICT["PROFILE_NAME"]
    )
    parser_detect_document_text.add_argument(
        "--region-name", type=str, help=STRING_DICT["REGION"]
    )
    parser_detect_document_text.add_argument(
        "--print",
        choices=[cd.name for cd in CLIPrint],
        nargs="+",
        help=STRING_DICT["PRINT"],
    )
    parser_detect_document_text.add_argument(
        "--linearize",
        action="store_true",
        help=STRING_DICT["LINEARIZE"],
    )
    parser_detect_document_text.add_argument(
        "--linearize-config-path",
        type=str,
        help=STRING_DICT["LINEARIZE_CONFIG"],
    )
    parser_detect_document_text.add_argument(
        "--overlay",
        choices=[cd.name for cd in CLIOverlay],
        nargs="+",
        help=STRING_DICT["OVERLAY"],
    )
    parser_detect_document_text.add_argument(
        "--font-size-ratio",
        type=float,
        help=STRING_DICT["FONT_SIZE_RATIO"],
        default=0.75,
    )

    parser_start_document_text_detection = subparsers.add_parser(
        "start-document-text-detection",
        help=STRING_DICT["START_DOCUMENT_TEXT_DETECTION"],
    )
    parser_start_document_text_detection.add_argument(
        "file_source", type=str, help=STRING_DICT["INPUT_FILE_ASYNC"]
    )
    parser_start_document_text_detection.add_argument(
        "--s3-upload-path", type=str, help=STRING_DICT["S3_UPLOAD_PATH"]
    )
    parser_start_document_text_detection.add_argument(
        "--s3-output-path", type=str, help=STRING_DICT["S3_OUTPUT_PATH"]
    )
    parser_start_document_text_detection.add_argument(
        "--profile-name", type=str, help=STRING_DICT["PROFILE_NAME"]
    )
    parser_start_document_text_detection.add_argument(
        "--region-name", type=str, help=STRING_DICT["REGION"]
    )

    # analyze-document parsers
    parser_analyze_document = subparsers.add_parser(
        "analyze-document", help=STRING_DICT["ANALYZE_DOCUMENT"]
    )
    parser_analyze_document.add_argument(
        "file_source", type=str, help=STRING_DICT["INPUT_FILE_SYNC"]
    )
    parser_analyze_document.add_argument(
        "output_file", type=str, help=STRING_DICT["OUTPUT_FILE"]
    )
    parser_analyze_document.add_argument(
        "--features",
        choices=[tf.name for tf in TextractFeatures],
        nargs="+",
        required=True,
    )
    parser_analyze_document.add_argument(
        "--queries", type=str, nargs="+", help=STRING_DICT["QUERIES"]
    )
    parser_analyze_document.add_argument(
        "--profile-name", type=str, help=STRING_DICT["PROFILE_NAME"]
    )
    parser_analyze_document.add_argument(
        "--region-name", type=str, help=STRING_DICT["REGION"]
    )
    parser_analyze_document.add_argument(
        "--print",
        choices=[cd.name for cd in CLIPrint],
        nargs="+",
        help=STRING_DICT["PRINT"],
    )
    parser_analyze_document.add_argument(
        "--linearize",
        action="store_true",
        help=STRING_DICT["LINEARIZE"],
    )
    parser_analyze_document.add_argument(
        "--linearize-config-path",
        type=str,
        help=STRING_DICT["LINEARIZE_CONFIG"],
    )
    parser_analyze_document.add_argument(
        "--overlay",
        choices=[cd.name for cd in CLIOverlay],
        nargs="+",
        help=STRING_DICT["OVERLAY"],
    )
    parser_analyze_document.add_argument(
        "--font-size-ratio",
        type=float,
        help=STRING_DICT["FONT_SIZE_RATIO"],
        default=0.75,
    )

    parser_start_document_analysis = subparsers.add_parser(
        "start-document-analysis", help=STRING_DICT["START_DOCUMENT_ANALYSIS"]
    )
    parser_start_document_analysis.add_argument(
        "file_source", type=str, help=STRING_DICT["INPUT_FILE_ASYNC"]
    )
    parser_start_document_analysis.add_argument(
        "--features",
        choices=[tf.name for tf in TextractFeatures],
        nargs="+",
        required=True,
    )
    parser_start_document_analysis.add_argument(
        "--queries", type=str, nargs="+", help=STRING_DICT["QUERIES"]
    )
    parser_start_document_analysis.add_argument(
        "--s3-upload-path", type=str, help=STRING_DICT["S3_UPLOAD_PATH"]
    )
    parser_start_document_analysis.add_argument(
        "--s3-output-path", type=str, help=STRING_DICT["S3_OUTPUT_PATH"]
    )
    parser_start_document_analysis.add_argument(
        "--profile-name", type=str, help=STRING_DICT["PROFILE_NAME"]
    )
    parser_start_document_analysis.add_argument(
        "--region-name", type=str, help=STRING_DICT["REGION"]
    )

    # analyze-expense parsers
    parser_analyze_expense = subparsers.add_parser(
        "analyze-expense", help=STRING_DICT["ANALYZE_EXPENSE"]
    )
    parser_analyze_expense.add_argument(
        "file_source", type=str, help=STRING_DICT["INPUT_FILE_SYNC"]
    )
    parser_analyze_expense.add_argument(
        "output_file", type=str, help=STRING_DICT["OUTPUT_FILE"]
    )
    parser_analyze_expense.add_argument(
        "--profile-name", type=str, help=STRING_DICT["PROFILE_NAME"]
    )
    parser_analyze_expense.add_argument(
        "--region-name", type=str, help=STRING_DICT["REGION"]
    )
    parser_analyze_expense.add_argument(
        "--print",
        choices=[cd.name for cd in CLIPrint],
        nargs="+",
        help=STRING_DICT["PRINT"],
    )
    parser_analyze_expense.add_argument(
        "--linearize",
        action="store_true",
        help=STRING_DICT["LINEARIZE"],
    )
    parser_analyze_expense.add_argument(
        "--linearize-config-path",
        type=str,
        help=STRING_DICT["LINEARIZE_CONFIG"],
    )
    parser_analyze_expense.add_argument(
        "--overlay",
        choices=[cd.name for cd in CLIOverlay],
        nargs="+",
        help=STRING_DICT["OVERLAY"],
    )
    parser_analyze_expense.add_argument(
        "--font-size-ratio",
        type=float,
        help=STRING_DICT["FONT_SIZE_RATIO"],
        default=0.75,
    )

    parser_start_expense_analysis = subparsers.add_parser(
        "start-expense-analysis", help=STRING_DICT["START_EXPENSE_ANALYSIS"]
    )
    parser_start_expense_analysis.add_argument(
        "file_source", type=str, help=STRING_DICT["INPUT_FILE_ASYNC"]
    )
    parser_start_expense_analysis.add_argument(
        "--s3-upload-path", type=str, help=STRING_DICT["S3_UPLOAD_PATH"]
    )
    parser_start_expense_analysis.add_argument(
        "--s3-output-path", type=str, help=STRING_DICT["S3_OUTPUT_PATH"]
    )
    parser_start_expense_analysis.add_argument(
        "--profile-name", type=str, help=STRING_DICT["PROFILE_NAME"]
    )
    parser_start_expense_analysis.add_argument(
        "--region-name", type=str, help=STRING_DICT["REGION"]
    )

    # analyze-id parsers
    parser_analyze_id = subparsers.add_parser(
        "analyze-id", help=STRING_DICT["ANALYZE_ID"]
    )
    parser_analyze_id.add_argument(
        "file_source", type=str, help=STRING_DICT["INPUT_FILE_SYNC"]
    )
    parser_analyze_id.add_argument(
        "output_file", type=str, help=STRING_DICT["OUTPUT_FILE"]
    )
    parser_analyze_id.add_argument(
        "--profile-name", type=str, help=STRING_DICT["PROFILE_NAME"]
    )
    parser_analyze_id.add_argument(
        "--region-name", type=str, help=STRING_DICT["REGION"]
    )
    parser_analyze_id.add_argument(
        "--print",
        choices=[cd.name for cd in CLIPrint],
        nargs="+",
        help=STRING_DICT["PRINT"],
    )
    # get-result parsers
    parser_get_result = subparsers.add_parser(
        "get-result", help=STRING_DICT["GET_RESULT"]
    )
    parser_get_result.add_argument(
        "job_id", type=str, help=STRING_DICT["GET_RESULT_JOB_ID"]
    )
    parser_get_result.add_argument(
        "api",
        type=str,
        choices=[ta.name for ta in TextractAPI],
        help=STRING_DICT["GET_RESULT_API"],
    )
    parser_get_result.add_argument(
        "output_file", type=str, help=STRING_DICT["OUTPUT_FILE"]
    )
    parser_get_result.add_argument(
        "--profile-name", type=str, help=STRING_DICT["PROFILE_NAME"]
    )
    parser_get_result.add_argument(
        "--region-name", type=str, help=STRING_DICT["REGION"]
    )
    parser_get_result.add_argument(
        "--print",
        choices=[cd.name for cd in CLIPrint],
        nargs="+",
        help=STRING_DICT["PRINT"],
    )
    parser_get_result.add_argument(
        "--linearize",
        action="store_true",
        help=STRING_DICT["LINEARIZE"],
    )
    parser_get_result.add_argument(
        "--linearize-config-path",
        type=str,
        help=STRING_DICT["LINEARIZE_CONFIG"],
    )
    return parser


def textractor_cli():
    """
    CLI for Textractor
    """

    parser = _build_parser()

    args = parser.parse_args()

    if args.subcommand is None:
        parser.print_help()
        return

    if (
        args.profile_name is None and
        args.region_name is None and
        (os.environ.get("AWS_REGION") or os.environ.get("AWS_DEFAULT_REGION"))
    ):
        args.region_name = os.environ.get("AWS_REGION") or os.environ.get("AWS_DEFAULT_REGION")
    elif args.profile_name is None and args.region_name is None:
        args.profile_name = "default"

    extractor = Textractor(
        profile_name=args.profile_name,
        region_name=args.region_name,
    )

    s3_output_path = None
    if "output_file" in args and args.output_file.startswith("s3://"):
        s3_output_path = args.output_file

    # ASYNC is handled differently
    if args.subcommand.startswith("start"):
        if not args.file_source.startswith("s3://") and args.s3_upload_path is None:
            raise InputError(
                "--s3-upload-path is required if file_source is not an S3 path"
            )
        if args.subcommand == "start-document-text-detection":
            out = extractor.start_document_text_detection(
                args.file_source,
                s3_output_path=args.s3_output_path,
                s3_upload_path=args.s3_upload_path,
                save_image=False,
            )
        elif args.subcommand == "start-document-analysis":
            out = extractor.start_document_analysis(
                args.file_source,
                features=[TextractFeatures[feat] for feat in args.features],
                queries=args.queries,
                s3_output_path=args.s3_output_path,
                s3_upload_path=args.s3_upload_path,
                save_image=False,
            )
        elif args.subcommand == "start-expense-analysis":
            out = extractor.start_expense_analysis(
                args.file_source,
                s3_output_path=args.s3_output_path,
                s3_upload_path=args.s3_upload_path,
                save_image=False,
            )
        print(out.job_id)
    # SYNC
    else:
        if args.subcommand == "detect-document-text":
            out = extractor.detect_document_text(
                args.file_source,
                s3_output_path=s3_output_path,
            )
        elif args.subcommand == "analyze-document":
            out = extractor.analyze_document(
                args.file_source,
                features=[TextractFeatures[feat] for feat in args.features],
                queries=args.queries,
                s3_output_path=s3_output_path,
            )
        elif args.subcommand == "analyze-expense":
            out = extractor.analyze_expense(
                args.file_source,
                s3_output_path=s3_output_path,
            )
        elif args.subcommand == "analyze-id":
            out = extractor.analyze_id(
                args.file_source,
            )
        elif args.subcommand == "get-result":
            out = extractor.get_result(
                args.job_id,
                TextractAPI[args.api],
            )

        if args.print is not None:
            if "ALL" in args.print or "TEXT" in args.print:
                print(out.text)
            if "ALL" in args.print or "TABLES" in args.print:
                print(out.tables.pretty_print())
            if "ALL" in args.print or "FORMS" in args.print:
                print(out.key_values.pretty_print())
            if "ALL" in args.print or "QUERIES" in args.print:
                print(out.queries.pretty_print())
            if "ALL" in args.print or "EXPENSES" in args.print:
                print(out.expense_documents.pretty_print())
            if "ALL" in args.print or "SIGNATURES" in args.print:
                print(out.signatures.pretty_print())
            if "ALL" in args.print or "IDS" in args.print:
                print(out.identity_documents.pretty_print())
            if "ALL" in args.print or "LAYOUTS" in args.print:
                print(out.layouts.pretty_print())

        if args.linearize is not None:
            if args.linearize_config_path is not None:
                with open(args.linearize_config_path, "r") as f:
                    config = yaml.safe_load(f)
                print(out.get_text(TextLinearizationConfig(**config)))
            else:
                print(out.get_text())

        if "overlay" in args and args.overlay is not None:
            entity_list = EntityList()
            if "ALL" in args.overlay or "WORDS" in args.overlay:
                entity_list += out.words
            if "ALL" in args.overlay or "LINES" in args.overlay:
                entity_list += out.lines
            if "ALL" in args.overlay or "TABLES" in args.overlay:
                entity_list += out.tables
            if "ALL" in args.overlay or "FORMS" in args.overlay:
                entity_list += out.key_values
            if "ALL" in args.overlay or "QUERIES" in args.overlay:
                entity_list += out.queries
            if "ALL" in args.overlay or "SIGNATURES" in args.overlay:
                entity_list += out.signatures
            if "ALL" in args.overlay or "LAYOUTS" in args.overlay:
                entity_list += out.layouts
            image = entity_list.visualize(
                with_text=True,
                font_size_ratio=args.font_size_ratio,
            )
            image.save(f"{args.output_file}.png")

        with open(args.output_file, "w") as f:
            json.dump(out.response, f)


if __name__ == "__main__":
    textractor_cli()



"""
    Define exceptions specific to textractor.
"""


class RegionMismatchError(Exception):
    """Raised when region on the profile_name does not match the region of the S3 bucket being accessed."""

    pass


class NoImageException(Exception):
    """Raised when visualize() method is called without saving image during Textract API call."""

    pass


class InputError(Exception):
    """Raised when function inputs are incorrect."""


class EntityListCreationError(Exception):
    """Raised when EntityList is created without passing any object or list of objects."""

    pass


class InvalidProfileNameError(Exception):
    """Raised when profile_name passed to Textractor is invalid."""

    pass


class S3FilePathMissing(Exception):
    """Raised when s3 file path is missing."""

    pass


class MissingDependencyException(Exception):
    """Raised when a dependency is missing for a specific code path"""

    pass


class IncorrectMethodException(Exception):
    """Raised when wrong endpoint is called."""

    pass


class UnhandledCaseException(Exception):
    """Raised when no statement matched the condition"""

    pass

class UnsupportedDocumentException(Exception):
    """Raised by the Textract API when the document could not be processed"""

    pass

class InvalidS3ObjectException(Exception):
    """Raised by the Textract API when an S3 object could not be accessed"""
    
    pass



"""The ExpenseDocument class is the object representation of an AnalyzeID response. It is similar to a dictionary. Despite its name it does not inherit from Document as the AnalyzeID response does not contains position information."""

from collections import defaultdict
from typing import List

from textractor.data.constants import (
    AnalyzeExpenseFieldsGroup as AEFieldsGroup,
    AnalyzeExpenseFields as AEFields,
)
from textractor.data.text_linearization_config import TextLinearizationConfig
from textractor.entities.expense_field import (
    ExpenseField,
    LineItemGroup,
    BoundingBox,
    DocumentEntity,
)


class Fields(dict):
    """
    Dictionary to hold Summary Fields
    Dynamically added properties to enable ease of discovery
    """

    def __init__(self):
        super(Fields, self).__init__()
        # We dynamically set the fields to None to help with discoverability
        for field in AEFields:
            setattr(
                self.__class__,
                field.name,
                property(lambda self, field=field: self.get(field.name)),
            )

    def __repr__(self):
        output = ""
        for key, value in self.items():
            output += f"{key}:"
            offset = 0
            if len(value):
                output += "\n"
                offset = 4
            for field in value:
                output += " " * offset + str(field).replace("\n", "\\n") + "\n"

        return output


class FieldsGroups(dict):
    """
    Summary Fields Group dictionary
    {GROUP_KEY_NAME: {GROUP_ID_1: [SUMMARY_FIELD1, SUMMARY_FIELD2]}}
    """

    def __init__(self):
        super(FieldsGroups, self).__init__()
        for group in AEFieldsGroup:
            setattr(
                self.__class__,
                group.name,
                property(lambda self, group=group: self.get(group.name)),
            )

    def __repr__(self):
        output = ""
        for key, group in self.items():
            output += f"{key}: \n"
            for block in group.values():
                for expense_field in block:
                    output += "  " + str(expense_field).replace("\n", "\\n") + "\n"
                output += "\n"
            output += "\n"
        return output

    def get_group_bboxes(self, key: str):
        """
        Return the enclosing bboxes for each group for a given group key
        :param key: Group key e.g VENDOR
        :return:
        """
        bboxes = []
        for groups in self.get(key, {}).values():
            bboxes.append(BoundingBox.enclosing_bbox([f.bbox for f in groups]))
        return bboxes


class ExpenseDocument(DocumentEntity):
    """
    Represents the description of a single expense document.
    """

    def __init__(
        self,
        summary_fields: List[ExpenseField],
        line_items_groups: List[LineItemGroup],
        bounding_box: BoundingBox,
        page: int,
    ):
        """
        :param summary_fields: List of ExpenseFields, not including line item ones
        :param line_items_groups: Groups of Line Item tables
        :param bounding_box: The bounding box for that ExpenseDocument
        :param page: The page where that document is
        """
        super().__init__("", bbox=bounding_box)
        self._summary_fields_list = summary_fields
        self._line_items_groups = line_items_groups
        self.summary_fields = Fields()
        self.summary_groups = FieldsGroups()
        self._unnormalized_fields = defaultdict(list)
        self._assign_summary_fields()
        self._page = page

    @property
    def page(self):
        return self._page

    @property
    def bbox(self):
        return BoundingBox.enclosing_bbox(
            [s.bbox for s in self._summary_fields_list]
            + [g.bbox for g in self._line_items_groups],
            spatial_object=self._bbox.spatial_object,
        )

    def _assign_summary_fields(self):
        for field in self._summary_fields_list:
            # We assign them as properties
            name = field.type.text

            # Adding it to the dicts of normalized field
            if name in self.summary_fields:
                self.summary_fields[name].append(field)
            else:
                self.summary_fields[name] = [field]

            # Adding it to the dicts of unnormalized fields using the provided key
            key = field.key.text if field.key else ""
            self._unnormalized_fields[key].append(field)

            # If the field is part of a group, we add it to the list of fields for that group
            for group_properties in field.group_properties:
                for property_type in group_properties.types:
                    if property_type not in self.summary_groups:
                        self.summary_groups[property_type] = dict()
                    if group_properties.id not in self.summary_groups[property_type]:
                        self.summary_groups[property_type][group_properties.id] = []
                    self.summary_groups[property_type][group_properties.id].append(
                        field
                    )

    @property
    def summary_fields_list(self):
        return self._summary_fields_list

    @property
    def line_items_groups(self) -> List[LineItemGroup]:
        return self._line_items_groups

    def get_text_and_words(
        self, config: TextLinearizationConfig = TextLinearizationConfig()
    ):
        # TODO
        return "", []

    def __repr__(self) -> str:
        output = f"Summary fields: {len(self.summary_fields)}\n"
        output += "Line Item Groups:"
        output += "\n" if len(self.line_items_groups) > 1 else " "
        for i, line_item in enumerate(self.line_items_groups):
            output += f"index {line_item.index}: {len(line_item.rows)} row{'s' if (len(line_item.rows) > 1) else ''}"
        return output



import time
import boto3
import os
import json
import datetime
from textractcaller.t_call import get_s3_output_config_keys, OutputConfig, remove_none


def results_exist(job_id: str, s3_bucket: str, s3_prefix: str, s3_client=None) -> bool:
    if not s3_client:
        s3_client = boto3.client("s3")
    response = s3_client.list_objects(
        Bucket=s3_bucket,
        Prefix=os.path.join(s3_prefix, job_id + "/"),
        Delimiter="/",
        MaxKeys=2,
    )
    # The directory will have at least one file because of the S3 access check
    return "Contents" in response and len(response["Contents"]) > 1

def get_full_json_from_output_config(
    output_config: OutputConfig, job_id: str, s3_client=None
) -> dict:
    if not output_config or not job_id:
        raise ValueError("no output_config or job_id")
    if not output_config.s3_bucket or not output_config.s3_prefix:
        raise ValueError("no output_config or job_id")
    if not s3_client:
        s3_client = boto3.client("s3")

    result_value = dict()
    last_result = None
    parsed_keys = set()
    while last_result is None or (datetime.datetime.now().astimezone() - last_result).total_seconds() < 5:
        keys = get_s3_output_config_keys(
            output_config=output_config, job_id=job_id, s3_client=s3_client
        )
        for key in keys:
            if key in parsed_keys:
                continue
            parsed_keys.add(key)
            s3_object = s3_client.get_object(Bucket=output_config.s3_bucket, Key=key)
            if last_result is None:
                last_result = s3_object["LastModified"]
            else:
                last_result = max(last_result, s3_object["LastModified"])
            body = s3_object["Body"]
            body_read = body.read()
            body_decode = body_read.decode("utf-8")
            response = dict(json.loads(body_decode))
            if "Blocks" in result_value:
                result_value["Blocks"].extend(response["Blocks"])
            else:
                result_value = response
    result_value = remove_none(result_value)
    return result_value



import dataclasses
import os
import logging
from textractor.data.text_linearization_config import TextLinearizationConfig

from textractor.entities.bbox import BoundingBox
from textractor.entities.document_entity import DocumentEntity
from textractor.data.constants import AnalyzeExpenseLineItemFields as AELineItems
from typing import List, Tuple


@dataclasses.dataclass
class ExpenseType:
    """
    Type of an ExpenseField, e.g TOTAL or SUBTOTAL
    """

    text: str
    confidence: float
    raw_object: object


@dataclasses.dataclass
class ExpenseGroupProperty:
    """
    Associated with a given ExpenseField, which group it is associated with and the related type of the group
    """

    id: str
    types: List[str]


class Expense(DocumentEntity):
    """
    Holds the Key or the Value of an Expense
    """

    def __init__(self, bbox: BoundingBox, text: str, confidence: float, page: int):
        super(Expense, self).__init__(entity_id="", bbox=bbox)
        self._text = text
        self._confidence = confidence
        self._bbox = bbox
        self._page = page

    @property
    def page(self):
        return self._page

    @property
    def text(self):
        return self._text

    @property
    def geometry(self):
        return self._bbox

    def __repr__(self):
        return self._text

    def get_text_and_words(
        self, config: TextLinearizationConfig = TextLinearizationConfig()
    ) -> Tuple[str, List]:
        """
        Used for linearization, returns the linearized text of the Expense and the matching words

        :return: Tuple of text and word list
        :rtype: Tuple[str, List[Word]]
        """
        return "", []


class ExpenseField(DocumentEntity):
    """
    The ExpenseField holds the information a given summary field, key, value and type.
    The bounding box of that ExpenseField is the enclosing one of all its components
    """

    def __init__(
        self,
        type: ExpenseType,
        value: Expense,
        group_properties: List[ExpenseGroupProperty],
        page: int,
        label: Expense = None,
        currency=None,
    ):
        enclosing_bbox = None
        if label:
            enclosing_bbox = BoundingBox.enclosing_bbox(
                [label.bbox, value.bbox], spatial_object=label.bbox.spatial_object
            )
        elif value.bbox:
            enclosing_bbox = BoundingBox.enclosing_bbox(
                [label, value], spatial_object=value.bbox.spatial_object
            )

        super(ExpenseField, self).__init__("", enclosing_bbox)
        self._enclosing_bbox = enclosing_bbox
        self._type = type
        self._key = label
        self._value = value
        self._group_properties = group_properties
        self._currency = currency
        self.raw_object = None
        self._page = page

    @property
    def bbox(self) -> BoundingBox:
        return self._enclosing_bbox

    @property
    def type(self) -> ExpenseType:
        return self._type

    @property
    def key(self) -> Expense:
        return self._key

    @property
    def value(self) -> Expense:
        return self._value

    @property
    def page(self) -> int:
        return self._page

    @property
    def group_properties(self) -> List[ExpenseGroupProperty]:
        return self._group_properties

    def __repr__(self) -> str:
        repr = ""

        if self.type:
            repr += f"{'_' if len(repr) else ''}{str(self.type.text)}"

        if self.key:
            repr += f" ({str(self.key.text)})"

        if self.value:
            repr += f": {str(self.value.text)}".replace("\n", "\\n")

        if self._currency:
            repr += f" [{self._currency}]"

        return repr

    def get_text_and_words(
        self, config: TextLinearizationConfig = TextLinearizationConfig()
    ) -> Tuple[str, List]:
        """
        Used for linearization, returns the linearized text of the ExpenseField and the matching words

        :return: Tuple of text and word list
        :rtype: Tuple[str, List[Word]]
        """
        return "", []


class LineItemRow(DocumentEntity):
    """
    A LineItemRow contains several ExpenseField that are all inside the row. They don't always align in
    a structured column structure as tables do.
    """

    def __init__(self, index, line_item_expense_fields: List[ExpenseField], page: int):
        super().__init__(
            "",
            bbox=BoundingBox.enclosing_bbox([i.bbox for i in line_item_expense_fields]),
        )
        self._index = index
        self._line_item_expense_fields = line_item_expense_fields
        self._page = page

    @property
    def page(self):
        return self._page

    @property
    def expenses(self):
        return self._line_item_expense_fields

    @property
    def bbox(self):
        # Dangerous, we need at least one expense in an expense row
        return BoundingBox.enclosing_bbox(
            [f.bbox for f in self._line_item_expense_fields],
            spatial_object=self.expenses[0].bbox.spatial_object,
        )

    def __getitem__(self, index):
        if isinstance(index, int):
            return self._line_item_expense_fields[index]
        else:
            for expense in self.expenses:
                if expense.type.text == index:
                    return expense
        raise IndexError(f"{index} is not present in this expense row")

    def get(self, index):
        return self[index]

    def __repr__(self):
        return str(self.expenses)

    def get_text_and_words(
        self, config: TextLinearizationConfig = TextLinearizationConfig()
    ) -> Tuple[str, List]:
        """
        Used for linearization, returns the linearized text of the LineItemRow and the matching words

        :return: Tuple of text and word list
        :rtype: Tuple[str, List[Word]]
        """
        return "", []


class LineItemGroup(DocumentEntity):
    """
    A LineItemGroup contains several LineItemRow. It is often similar to a table in invoices
    but in receipts, the table structure can be more loose and less aligned.
    """

    def __init__(self, index, line_item_rows: List[LineItemRow], page: int):
        super(LineItemGroup, self).__init__(
            "", BoundingBox.enclosing_bbox([f.bbox for f in line_item_rows])
        )
        self._index = index
        self._line_item_rows = line_item_rows
        self._page = page

    @property
    def page(self):
        return self._page

    @property
    def rows(self):
        return self._line_item_rows

    @property
    def index(self):
        return self._index

    @property
    def bbox(self):
        # Dangerous, we need at least one line item in a line item rows
        return BoundingBox.enclosing_bbox(
            [f.bbox for f in self._line_item_rows],
            self._line_item_rows[0].bbox.spatial_object,
        )

    def to_pandas(self, include_EXPENSE_ROW=False):
        try:
            from pandas import DataFrame
        except ImportError:
            logging.info(
                "pandas library is required for exporting tables to DataFrame objects"
            )
            return None

        types = {field.name for field in AELineItems}
        columns = {field.name: i for i, field in enumerate(AELineItems)}
        for row in self.rows:
            for expense in row.expenses:
                if expense.type.text not in columns:
                    columns[expense.type.text] = max(columns.values()) + 1
                    types.add(expense.type.text)
        data = [["" for _ in range(len(types))] for _ in range(len(self.rows))]

        for i, row in enumerate(self.rows):
            for j, expense in enumerate(row.expenses):
                data[i][columns[expense.type.text]] = expense.value.text

        df = DataFrame(data=data, columns=columns.keys())

        if not include_EXPENSE_ROW:
            df = df.drop(columns=[AELineItems.EXPENSE_ROW.name])

        return df

    def to_csv(self):
        return self.to_pandas().to_csv()

    def to_json(self):
        return self.to_pandas().to_json()

    def __repr__(self) -> str:
        output = ""
        for row in self.rows:
            output += "|"
            for expense in row:
                output += (
                    expense.type.text
                    + ": "
                    + expense.value.text.replace(os.linesep, "\\n")
                )
                output += " | "
            output += os.linesep
        return output

    def __getitem__(self, index):
        return self._line_item_rows[index]

    def get_text_and_words(
        self, config: TextLinearizationConfig = TextLinearizationConfig()
    ) -> Tuple[str, List]:
        """
        Used for linearization, returns the linearized text of the LineItemGroup and the matching words

        :return: Tuple of text and word list
        :rtype: Tuple[str, List[Word]]
        """
        return "", []



"""
Consumes Textract JSON response and converts them to a Document object format.
This class contains all the necessary utilities to create entity objects from JSON blocks within the response.
Use ResponseParser's parse function to handle API response and convert them to Document objects.
"""

import logging
import uuid
from copy import deepcopy
from typing import Any, List, Dict, Tuple
from collections import defaultdict
from textractor.entities.identity_document import IdentityDocument
from textractor.entities.expense_document import ExpenseDocument
from textractor.entities.expense_field import (
    Expense,
    ExpenseField,
    ExpenseType,
    ExpenseGroupProperty,
    LineItemGroup,
    LineItemRow,
)

from textractor.entities.page import Page
from textractor.entities.query_result import QueryResult
from textractor.entities.signature import Signature
from textractor.entities.word import Word
from textractor.entities.line import Line
from textractor.entities.value import Value
from textractor.entities.table import Table
from textractor.entities.bbox import BoundingBox
from textractor.entities.document import Document
from textractor.entities.key_value import KeyValue
from textractor.entities.table_cell import TableCell
from textractor.entities.table_title import TableTitle
from textractor.entities.table_footer import TableFooter
from textractor.entities.query import Query
from textractor.entities.selection_element import SelectionElement
from textractor.entities.layout import Layout
from textractor.data.constants import (
    LAYOUT_ENTITY,
    LAYOUT_FIGURE,
    TABLE_FOOTER,
    TABLE_TITLE,
    COLUMN_HEADER,
    TABLE_SUMMARY,
    TABLE_SECTION_TITLE,
    TABLE_STRUCTURED,
    TABLE_SEMI_STRUCTURED,
    SelectionStatus,
    TextTypes,
    TableTypes,
    HANDWRITING,
    PRINTED,
    WORD,
    LINE,
    KEY_VALUE_SET,
    CELL,
    TABLE,
    SELECTION_ELEMENT,
    PAGE,
    MERGED_CELL,
    QUERY,
    SIGNATURE,
    LAYOUT,
    LAYOUT_LIST,
    LAYOUT_TABLE,
    LAYOUT_KEY_VALUE,
)

THRESHOLD = 0.95


def _create_document_object(response: dict) -> Document:
    """
    Consumes API Response in JSON format and creates a Document object.

    :param response: json response from Textract API
    :type response: dict

    :return: Returns a Document object populated with metadata on number of pages.
    :rtype: Document
    """
    doc = Document(num_pages=response["DocumentMetadata"]["Pages"])
    return doc


def _filter_block_type(response: dict, entity: str) -> List[Dict[str, Any]]:
    """
    Consumes entire JSON response, filters and returns list of blocks corresponding to the entity
    parameter from API response JSON.

    :param response: JSON response from Textract API
    :type response: dict
    :param entity: Entity to be extracted from the JSON response
    :type entity: str

    :return: Returns a list of JSON blocks that match entity parameter.
    :rtype: List
    """
    return [block for block in response["Blocks"] if block["BlockType"] == entity]


def _filter_by_entity(
    block_json: List[Dict[str, Any]], entity_type: str
) -> Dict[str, Any]:
    """
    Filters and returns dictionary of blocks corresponding to the entity_type from API response JSON.

    :param block_json: list of blocks belonging to a specific entity
    :type block_json: List[Dict[str, Any]]
    :param entity_type: EntityType used to select/filter from list of blocks
    :type entity_type: str

    :return: Dictionary mapping of block ID with JSON block for entity type.
    :rtype: Dict[str, Any]
    """
    return {
        block["Id"]: block
        for block in block_json
        if (
            "EntityTypes" in block and
            len(block["EntityTypes"]) and
            block["EntityTypes"][0] == entity_type
        )
    }


def _get_relationship_ids(block_json: Dict[str, Any], relationship: str) -> List[str]:
    """
    Takes the JSON block corresponding to an entity and returns the Ids of the chosen Relationship if the Relationship exists.

    :param block_json: JSON block corresponding to an entity
    :type block_json: List[Dict[str, Any]]
    :relationship: CHILD or VALUE as input
    :type relationship: str

    :return: List of IDs with type Relationship to entity
    :rtype: List
    """
    ids = []
    try:
        ids = [
            rel["Ids"]
            for rel in block_json["Relationships"]
            if rel["Type"] == relationship
        ][0]
    except:
        logging.info(
            f"{block_json['BlockType']} - {block_json['Id']} does not have ids with {relationship} relationship."
        )
    return ids


def _create_page_objects(
    response: dict,
) -> Tuple[Dict[str, Page], List[Dict[str, Any]]]:
    """
    Consumes API Response in JSON format and returns Page objects for the Document.

    :param response: JSON response from Textract API
    :type response: dict

    :return: Returns dictionary with page ID - Page object mapping,  list of JSON blocks belonging to PAGE blocks.
    :rtype: Dict[str, Page], List[str]
    """
    pages = []
    page_elements = _filter_block_type(response, entity=PAGE)

    for page_json in page_elements:
        asset_id = page_json["Id"]
        width = page_json["Geometry"]["BoundingBox"]["Width"]
        height = page_json["Geometry"]["BoundingBox"]["Height"]
        page_num = page_json["Page"] if len(page_elements) > 1 else 1
        page_children = _get_relationship_ids(page_json, relationship="CHILD")
        page = Page(
            id=asset_id,
            width=width,
            height=height,
            page_num=page_num,
            child_ids=page_children,
        )
        pages.append(page)
    pages = {page.id: page for page in pages}
    return pages, page_elements


def _create_word_objects(
    word_ids: List[str],
    id_json_map: Dict[str, str],
    existing_words: Dict[str, Word],
    page: Page,
) -> List[Word]:
    """
    Creates list of Word objects for all word_ids passed to the function.

    :param word_ids: List of ids corresponding to the words present within Page.
    :type word_ids: list
    :param id_json_map: Dictionary containing entity_id: JSON block mapping.
    :type id_json_map: dict
    :param page: Instance of parent Page object.
    :type page: Page

    :return: Returns a list of Word objects for the IDs passed in word_ids.
    :rtype: list
    """
    words = []
    text_type = {PRINTED: TextTypes.PRINTED, HANDWRITING: TextTypes.HANDWRITING}

    for word_id in word_ids:
        if word_id in existing_words:
            words.append(existing_words[word_id])
        else:
            # FIXME: This could be gated
            if not word_id in id_json_map:
                continue
            elem = id_json_map[word_id]
            word = Word(
                entity_id=elem["Id"],
                bbox=BoundingBox.from_normalized_dict(
                    elem["Geometry"]["BoundingBox"], spatial_object=page
                ),
                text=elem.get("Text"),
                text_type=text_type[elem.get("TextType")],
                confidence=elem["Confidence"],
            )
            word.raw_object = elem
            words.append(word)
            existing_words[word_id] = word

    for word in words:
        word.page = page.page_num
        word.page_id = page.id

    return words


def _create_line_objects(
    line_ids: List[str],
    id_json_map: Dict[str, str],
    existing_words: Dict[str, Word],
    page: Page,
) -> Tuple[List[Line], List[Word]]:
    """
    Creates list of Line objects for all lines in the Page derived from the API JSON response.

    :param line_ids: List of IDs corresponding to the lines present within Page.
    :type line_ids: list
    :param id_json_map: Dictionary containing entity_id: JSON block mapping.
    :type id_json_map: dict
    :param page: Instance of parent Page object.
    :type page: Page

    :return: Returns a list of Line objects for the IDs passed in line_ids and list of Word objects
             belonging to the corresponding Line objects.
    :rtype: List[Line], List[Word]
    """
    page_lines = []

    for line_id in line_ids:
        if line_id in page.child_ids:
            page_lines.append(id_json_map[line_id])

    lines = []
    page_words = []
    for line in page_lines:
        if _get_relationship_ids(line, relationship="CHILD"):
            line_words = _create_word_objects(
                _get_relationship_ids(line, relationship="CHILD"),
                id_json_map,
                existing_words,
                page,
            )
            page_words.extend(line_words)
            lines.append(
                Line(
                    entity_id=line["Id"],
                    bbox=BoundingBox.from_normalized_dict(
                        line["Geometry"]["BoundingBox"], spatial_object=page
                    ),
                    words=line_words,
                    confidence=line["Confidence"],
                )
            )
            for word in line_words:
                word.line = lines[-1]
                word.line_id = lines[-1].id
                word.line_bbox = lines[-1].bbox
            lines[-1]._children = line_words
            lines[-1].raw_object = line

    for line in lines:
        line.page = page.page_num
        line.page_id = page.id

    return lines, page_words


def _create_selection_objects(
    selection_ids: List[str], id_json_map: Dict[str, Any], page: Page
) -> Dict[str, SelectionElement]:
    """
    Creates dictionary mapping of SelectionElement ID with SelectionElement objects for all ids passed in selection_ids.

    :param selection_ids: List of ids corresponding to the SelectionElements.
    :type selection_ids: list
    :param id_json_map: Dictionary containing entity_id: JSON block mapping.
    :type id_json_map: dict
    :param page: Instance of parent Page object.
    :type page: Page

    :return: Returns a dictionary mapping of SelectionElement IDs with SelectionElement objects for the IDs present in
             selection_ids.
    :rtype: Dict[str, SelectionElement]
    """
    checkbox_elements = [id_json_map[selection_id] for selection_id in selection_ids]

    status = {
        "SELECTED": SelectionStatus.SELECTED,
        "NOT_SELECTED": SelectionStatus.NOT_SELECTED,
    }

    checkboxes = {}
    for block in checkbox_elements:
        checkboxes[block["Id"]] = SelectionElement(
            entity_id=block["Id"],
            bbox=BoundingBox.from_normalized_dict(
                block["Geometry"]["BoundingBox"], spatial_object=page
            ),
            status=status[block["SelectionStatus"]],
            confidence=block["Confidence"],
        )
        checkboxes[block["Id"]].raw_object = block

    for c in checkboxes.values():
        c.page = page.page_num
        c.page_id = page.id

    return checkboxes


def _create_value_objects(
    value_ids: List[str],
    id_json_map: Dict[str, Any],
    entity_id_map: Dict[str, list],
    existing_words: Dict[str, Word],
    page: Page,
) -> Tuple[Dict[str, Value], Dict[str, SelectionElement]]:
    """
    Creates dictionary containing Value objects for all value_ids in the Page derived from the API response JSON.

    :param value_ids: List of ids corresponding to the Values in the page.
    :type value_ids: list
    :param id_json_map: Dictionary containing entity_id:JSON block mapping.
    :type id_json_map: dict
    :param entity_id_map: Dictionary containing entity_type:List[entity_id] mapping.
    :type entity_id_map: dict
    :param page: Instance of parent Page object.
    :type page: Page

    :return: Dictionary mapping value_ids to Value objects.
    :rtype: Dict[str, Value]
    """
    # FIXME: This could be gated
    values_info = {value_id: id_json_map.get(value_id, None) for value_id in value_ids}

    values = {}
    for block_id, block in values_info.items():
        # FIXME: This should be gated
        if block is None:
            continue
        values[block_id] = Value(
            entity_id=block_id,
            bbox=BoundingBox.from_normalized_dict(
                block["Geometry"]["BoundingBox"], spatial_object=page
            ),
            confidence=block["Confidence"],
        )
        values[block_id].raw_object = block

    checkboxes = _create_selection_objects(
        entity_id_map[SELECTION_ELEMENT], id_json_map, page
    )

    # Add children to Value object
    for val_id in values.keys():
        val_child_ids = _get_relationship_ids(values_info[val_id], relationship="CHILD")
        for child_id in val_child_ids:
            # FIXME: This should be gated
            if child_id not in id_json_map:
                continue
            if id_json_map[child_id]["BlockType"] == WORD:
                words = _create_word_objects(
                    [child_id], id_json_map, existing_words, page
                )
                values[val_id].words += words
                values[val_id].add_children(words)

            elif id_json_map[child_id]["BlockType"] == SIGNATURE:
                continue
            else:
                checkbox = checkboxes[child_id]
                checkbox.value_id = val_id
                values[val_id].add_children([checkbox])
                values[val_id].contains_checkbox = True

            values[val_id].page = page.page_num
            values[val_id].page_id = page.id

    return values, checkboxes


def _create_query_objects(
    query_ids: List[str],
    id_json_map: Dict[str, str],
    entity_id_map: Dict[str, list],
    page: Page,
) -> List[Query]:
    page_queries = []
    for query_id in query_ids:
        if query_id in page.child_ids:
            page_queries.append(id_json_map[query_id])

    query_result_id_map = {}
    for block in page_queries:
        answer = _get_relationship_ids(block, relationship="ANSWER")
        query_result_id_map[block["Id"]] = answer[0] if answer else None

    query_results = _create_query_result_objects(
        list(query_result_id_map.values()), id_json_map, entity_id_map, page
    )

    queries = []
    for query in page_queries:
        query_result = query_results.get(query_result_id_map[query["Id"]])
        query_obj = Query(
            query["Id"],
            query["Query"]["Text"],
            query["Query"].get("Alias"),
            query_result,
            query_result.bbox if query_result is not None else None,
        )
        query_obj.raw_object = query
        queries.append(query_obj)

    return queries


def _create_query_result_objects(
    query_result_ids: List[str],
    id_json_map: Dict[str, str],
    entity_id_map: Dict[str, list],
    page: Page,
) -> Dict[str, QueryResult]:
    page_query_results = []
    for query_result_id in query_result_ids:
        if query_result_id in page.child_ids and query_result_id in id_json_map:
            page_query_results.append(id_json_map[query_result_id])

    query_results = {}
    for block in page_query_results:
        query_results[block["Id"]] = QueryResult(
            entity_id=block["Id"],
            confidence=block["Confidence"],
            result_bbox=BoundingBox.from_normalized_dict(
                (
                    block.get("Geometry") or
                    {
                        "BoundingBox": {
                            "Width": 1.0,
                            "Height": 1.0,
                            "Left": 0.0,
                            "Top": 0.0,
                        }
                    }
                )["BoundingBox"],
                spatial_object=page,
            ),
            answer=block["Text"],
        )
        query_results[block["Id"]].raw_object = block

    for query_result_id, query_result in query_results.items():
        query_result.page = page.page_num
        query_result.page_id = page.id

    return query_results


def _create_signature_objects(
    signature_ids: List[str],
    id_json_map: Dict[str, str],
    entity_id_map: Dict[str, list],
    page: Page,
) -> Dict[str, Signature]:
    page_signatures = []
    for signature_id in signature_ids:
        if signature_id in page.child_ids:
            page_signatures.append(id_json_map[signature_id])

    signatures = {}
    for block in page_signatures:
        signatures[block["Id"]] = Signature(
            entity_id=block["Id"],
            confidence=block["Confidence"],
            bbox=BoundingBox.from_normalized_dict(
                block["Geometry"]["BoundingBox"], spatial_object=page
            ),
        )
        signatures[block["Id"]].raw_object = block

    for signature_id, signature in signatures.items():
        signature.page = page.page_num
        signature.page_id = page.id

    signatures_added = set()
    for layout in sorted(page.leaf_layouts, key=lambda x: x.bbox.y):
        if layout.layout_type == LAYOUT_ENTITY:
            continue
        for signature in sorted(signatures.values(), key=lambda x: x.bbox.y):
            if (
                layout.bbox.get_intersection(signature.bbox).area
                > THRESHOLD * signature.bbox.area
                and signature not in signatures_added
            ):
                layout.children.append(signature)
                signatures_added.add(signature)
                del signatures[signature.id]

    signature_layouts = []
    for signature in signatures.values():
        if signature not in signatures_added:
            signatures_added.add(signature)
            layout = Layout(
                entity_id=signature.id,
                bbox=signature.bbox,
                label=LAYOUT_ENTITY,
                reading_order=-1,
            )
            layout.children.append(signature)
            layout.page = page.page_num
            layout.page_id = page.id
            signature_layouts.append(layout)

    layouts_to_remove = []
    for layout in page.leaf_layouts:
        layouts_that_intersect = []
        for signature_layout in signature_layouts:
            intersection = layout.bbox.get_intersection(signature_layout.bbox).area
            if intersection:
                layouts_that_intersect.append(signature_layout)
        words_in_sub_layouts = set()
        for i, intersect_layout in enumerate(
            sorted(layouts_that_intersect, key=lambda l: (l.bbox.y, l.bbox.x))
        ):
            intersect_layout.reading_order = (
                (layout.reading_order + (i + 1) * 0.1)
                if intersect_layout.reading_order == -1
                else min(
                    intersect_layout.reading_order, layout.reading_order + (i + 1) * 0.1
                )
            )
            for w in intersect_layout.children[0].words:
                words_in_sub_layouts.add(w.id)
        if words_in_sub_layouts:
            remaining_words = []
            for w in layout.words:
                if w.id not in words_in_sub_layouts:
                    remaining_words.append(w)
            if remaining_words:
                layout.bbox = BoundingBox.enclosing_bbox(
                    [w.bbox for w in remaining_words]
                )
                layout._children = list(set([w.line for w in remaining_words]))
            else:
                layouts_to_remove.append(layout)

    for layout in layouts_to_remove:
        page.leaf_layouts.remove(layout)

    for layout in signature_layouts:
        page.leaf_layouts.append(layout)

    return list(signatures_added)



def _create_keyvalue_objects(
    key_value_ids: List[str],
    id_json_map: Dict[str, Any],
    id_entity_map: Dict[str, str],
    entity_id_map: Dict[str, list],
    existing_words: Dict[str, Word],
    page: Page,
) -> Tuple[List[KeyValue], List[Word], Dict[str, SelectionElement]]:
    """
    Creates list of KeyValue objects for all key-value pairs in the Page derived from the API response JSON.

    :param key_value_ids: List of ids corresponding to the KeyValues in the page.
    :type key_value_ids: list
    :param id_json_map: Dictionary containing entity_id:JSON block mapping.
    :type id_json_map: dict
    :param entity_id_map: Dictionary containing entity_type:List[entity_id] mapping.
    :type entity_id_map: dict
    :param id_entity_map: Dictionary containing entity_id:entity_type mapping.
    :type id_entity_map: dict
    :param page: Instance of parent Page object.
    :type page: Page

    :return: Returns a list of KeyValue objects and list of Word objects with CHILD relationship
             to the KeyValue objects.
    :rtype: List[KeyValue], List[Word]
    """
    page_kv = []
    for kv_id in key_value_ids:
        if kv_id in page.child_ids:
            page_kv.append(id_json_map[kv_id])

    keys_info = _filter_by_entity(page_kv, entity_type="KEY")

    key_value_id_map = {
        block["Id"]: _get_relationship_ids(block, relationship="VALUE")[0]
        for block in keys_info.values()
    }

    values, selection_elements = _create_value_objects(
        list(key_value_id_map.values()),
        id_json_map,
        entity_id_map,
        existing_words,
        page,
    )

    keys = {}
    for block in keys_info.values():
        keys[block["Id"]] = KeyValue(
            entity_id=block["Id"],
            bbox=BoundingBox.from_normalized_dict(
                block["Geometry"]["BoundingBox"], spatial_object=page
            ),
            # FIXME: Should be gated
            contains_checkbox=(
                key_value_id_map[block["Id"]] in values and
                values[key_value_id_map[block["Id"]]].contains_checkbox
            ),
            value=values.get(key_value_id_map[block["Id"]]),
            confidence=block["Confidence"],
        )
        keys[block["Id"]].raw_object = block

    # Add words and children (Value Object) to KeyValue object
    kv_words = []
    for key_id in keys.keys():
        if keys[key_id].value is None:
            continue
        keys[key_id].value.key_id = key_id
        if keys[key_id].contains_checkbox:
            keys[key_id].value.children[0].key_id = key_id
            keys[key_id].selection_status = [
                c
                for c in keys[key_id].value.children
                if c.__class__.__name__ == "SelectionElement"
            ][0].status
        else:
            kv_words.extend(values[key_value_id_map[key_id]].words)

        key_child_ids = _get_relationship_ids(keys_info[key_id], relationship="CHILD")
        key_word_ids = [
            child_id
            for child_id in key_child_ids
            # FIXME: This should be gated
            if child_id in id_json_map and id_json_map[child_id]["BlockType"] == WORD
        ]
        key_words = _create_word_objects(
            key_word_ids, id_json_map, existing_words, page
        )

        key_child_ids = [
            child_id for child_id in key_child_ids if child_id not in key_word_ids
        ]
        # find a place to add selection elements for keys

        keys[key_id].key = key_words
        keys[key_id].add_children(key_words)
        keys[key_id].add_children([keys[key_id].value])
        kv_words.extend(key_words)

    key_values = list(keys.values())
    for kv in key_values:
        kv.bbox = BoundingBox.enclosing_bbox([kv.bbox] + ([kv.value.bbox] if kv.value is not None else []))
        kv.page = page.page_num
        kv.page_id = page.id

    #kv_added = set()
    #for layout in sorted(page.leaf_layouts, key=lambda x: x.bbox.y):
    #    if layout.layout_type == LAYOUT_ENTITY:
    #        continue
    #    for kv in sorted(key_values, key=lambda x: x.bbox.y):
    #        if (
    #            layout.bbox.get_intersection(kv.bbox).area > THRESHOLD * kv.bbox.area
    #            and kv not in kv_added
    #        ):
    #            layout.children.append(kv)
    #            kv_added.add(kv)
    #            key_values.remove(kv)

    return key_values, kv_words, selection_elements


def _create_layout_objects(
    layout_ids: List[Any],
    id_json_map: Dict[str, str],
    id_entity_map: Dict[str, List[str]],
    line_by_id: Dict[str, Line],
    page: Page,
) -> Tuple[List[Layout], List[Layout]]:
    """
    Creates Layout objects.

    :param page_layouts: Reading-ordered list containing JSON structure of tables within the page.
    :type page_layouts: list
    :param id_json_map: Dictionary containing entity_id:JSON block mapping.
    :type id_json_map: dict
    :param id_entity_map: Dictionary containing entity_id:entity_type mapping.
    :type id_entity_map: dict
    :param page: Instance of parent Page object.
    :type page: Page

    :return: Returns a list containing Layout objects.
    :rtype: List[Layout]
    """

    page_layouts = []
    for layout_id in layout_ids:
        if layout_id in page.child_ids:
            page_layouts.append(id_json_map[layout_id])

    leaf_layouts = []
    container_layouts = []
    parsed_blocks = set()
    for i, block in enumerate(page_layouts):
        if block["Id"] in parsed_blocks:
            continue
        if block["BlockType"] in (LAYOUT_LIST,):
            container_layouts.append(
                Layout(
                    entity_id=block["Id"],
                    confidence=block["Confidence"],
                    reading_order=i,
                    label=block["BlockType"],
                    bbox=BoundingBox.from_normalized_dict(
                        block["Geometry"]["BoundingBox"], spatial_object=page
                    ),
                )
            )
            parsed_blocks.add(block["Id"])
            for relationship in (block.get("Relationships", []) or []):
                if relationship["Type"] != "CHILD":
                    continue
                for leaf_id in relationship["Ids"]:
                    block = id_json_map[leaf_id]
                    parsed_blocks.add(leaf_id)
                    container_layouts[-1].children.append(
                        Layout(
                            entity_id=block["Id"],
                            confidence=block["Confidence"],
                            reading_order=i,
                            label=block["BlockType"],
                            bbox=BoundingBox.from_normalized_dict(
                                block["Geometry"]["BoundingBox"], spatial_object=page
                            ),
                        )
                    )
                    container_layouts[-1].children[-1].raw_object = block
                    for relationship in (block.get("Relationships", []) or []):
                        if relationship["Type"] != "CHILD":
                            continue
                        container_layouts[-1].children[-1].add_children(
                            [line_by_id[line_id] for line_id in relationship["Ids"] if line_id in line_by_id]
                        )
        else:
            leaf_layouts.append(
                Layout(
                    entity_id=block["Id"],
                    confidence=block["Confidence"],
                    reading_order=i,
                    label=block["BlockType"],
                    bbox=BoundingBox.from_normalized_dict(
                        block["Geometry"]["BoundingBox"], spatial_object=page
                    ),
                )
            )
            leaf_layouts[-1].raw_object = block
            for relationship in (block.get("Relationships", []) or []):
                if relationship["Type"] != "CHILD":
                    continue
                leaf_layouts[-1].add_children(
                    [line_by_id[line_id] for line_id in relationship["Ids"] if line_id in line_by_id]
                )

    for layout in leaf_layouts + container_layouts:
        layout.page = page.page_num
        layout.page_id = page.id

    return container_layouts, leaf_layouts


def _create_table_cell_objects(
    page_tables: List[Any],
    id_entity_map: Dict[str, List[str]],
    id_json_map: Dict[str, str],
    page: Page,
) -> Tuple[Dict[str, TableCell], Dict[str, Any]]:
    """
    Creates TableCell objects for all page_tables passed as input present on a single Page of the Document.

    :param page_tables: List containing JSON structure of tables within the page.
    :type page_tables: list
    :param id_entity_map: Dictionary containing entity_id:entity_type mapping.
    :type id_entity_map: dict
    :param id_json_map: Dictionary containing entity_id:JSON block mapping.
    :type id_json_map: dict
    :param page: Instance of parent Page object.
    :type page: Page

    :return: Returns a dictionary containing TableCells mapped with their IDs and dictionary containing ID: CELL JSON mapping.
    :rtype: Dict[str, TableCell], Dict[str, Any]
    """
    all_table_cells_info = {}
    for table in page_tables:
        for cell_id in _get_relationship_ids(table, relationship="CHILD"):
            # FIXME: This should be gated
            if cell_id in id_entity_map and id_entity_map[cell_id] == CELL:
                all_table_cells_info[cell_id] = id_json_map[cell_id]

    table_cells = {}
    for elem_id, elem in all_table_cells_info.items():
        entity_types = elem.get("EntityTypes", []) or []
        table_cells[elem_id] = TableCell(
            entity_id=elem_id,
            bbox=BoundingBox.from_normalized_dict(
                elem["Geometry"]["BoundingBox"], spatial_object=page
            ),
            row_index=elem["RowIndex"],
            col_index=elem["ColumnIndex"],
            row_span=elem["RowSpan"],
            col_span=elem["ColumnSpan"],
            confidence=elem["Confidence"],
            is_column_header=COLUMN_HEADER in entity_types,
            is_title=TABLE_TITLE in entity_types,
            is_footer=TABLE_FOOTER in entity_types,
            is_summary=TABLE_SUMMARY in entity_types,
            is_section_title=TABLE_SECTION_TITLE in entity_types,
        )
        table_cells[elem_id].raw_object = elem

    for cell in table_cells.values():
        cell.page = page.page_num
        cell.page_id = page.id
    return table_cells, all_table_cells_info


def _create_table_objects(
    table_ids: List[str],
    id_json_map: Dict[str, Any],
    id_entity_map: Dict[str, List[str]],
    entity_id_map: Dict[str, List[str]],
    existing_words: Dict[str, Word],
    key_values: Dict[str, KeyValue],
    checkboxes: Dict[str, SelectionElement],
    page: Page,
) -> Tuple[List[Table], List[Word]]:
    """
    Creates list of Table objects for all tables in the Page derived from the API response JSON.
    This includes creating TableCell objects and updating metadata for each cell. The TableCell objects are assigned as children
    of the table.

    :param table_ids: List of ids corresponding to the Tables in the page.
    :type table_ids: list
    :param id_json_map: Dictionary containing entity_id:JSON block mapping.
    :type id_json_map: dict
    :param id_entity_map: Dictionary containing entity_id:entity_type mapping.
    :type id_entity_map: dict
    :param entity_id_map: Dictionary containing entity_type:List[entity_id] mapping.
    :type entity_id_map: dict
    :param page: Instance of parent Page object.
    :type page: Page

    :return: Returns a list of table objects and list of words present in tables.
    :rtype: List[Table], List[Word]
    """
    # Create Tables
    page_tables = []
    for table_id in table_ids:
        if table_id in page.child_ids:
            page_tables.append(id_json_map[table_id])

    tables = {}
    for val in page_tables:
        tables[val["Id"]] = Table(
            entity_id=val["Id"],
            bbox=BoundingBox.from_normalized_dict(
                val["Geometry"]["BoundingBox"], spatial_object=page
            ),
        )
        # Setting table type based on the entity types present in the table
        if TABLE_STRUCTURED in (val.get("EntityTypes", []) or []):
            tables[val["Id"]].table_type = TableTypes.STRUCTURED
        elif TABLE_SEMI_STRUCTURED in (val.get("EntityTypes", []) or []):
            tables[val["Id"]].table_type = TableTypes.SEMI_STRUCTURED
        else:
            tables[val["Id"]].table_type = TableTypes.UNKNOWN
        # Setting raw JSON in the resulting object
        tables[val["Id"]].raw_object = val

    # Create Table Cells
    table_cells, all_table_cells_info = _create_table_cell_objects(
        page_tables, id_entity_map, id_json_map, page
    )

    merged_table_cells = [
        id_json_map[merge_id] for merge_id in entity_id_map[MERGED_CELL]
    ]

    # Add children to cells
    merged_child_map = {
        merged_cell["Id"]: _get_relationship_ids(merged_cell, relationship="CHILD")
        for merged_cell in merged_table_cells
    }
    merged_child_ids = sum([ids for ids in merged_child_map.values()], [])

    table_words = []
    added_key_values = set()
    for cell_id, cell in all_table_cells_info.items():
        children = _get_relationship_ids(cell, relationship="CHILD")
        # FIXME: This should be gated
        cell_word_ids = [
            child_id for child_id in children if (child_id in id_entity_map and id_entity_map[child_id] == WORD)
        ]
        # FIXME: This should be gated
        selection_ids = [
            child_id
            for child_id in children
            if child_id in id_entity_map and id_entity_map[child_id] == SELECTION_ELEMENT
        ]

        cell_words = _create_word_objects(
            cell_word_ids, id_json_map, existing_words, page
        )
        for w in cell_words:
            w.cell_id = table_cells[cell_id].id
            w.cell_bbox = table_cells[cell_id].bbox
            w.row_span = table_cells[cell_id].row_span
            w.col_span = table_cells[cell_id].col_span
            w.row_index = table_cells[cell_id].row_index
            w.col_index = table_cells[cell_id].col_index
        table_words.extend(cell_words)

        table_cells[cell_id].add_children(cell_words)
        
        # There are two types of selection elements, one comes from Tables, the other for KVs
        # This tries to reconcile both and to insert the selection element in the right place
        for child_id in selection_ids:
            if checkboxes[child_id].key_id in added_key_values:
                continue
            # This is a KeyValue
            if checkboxes[child_id].key_id is not None:
                kv = key_values[checkboxes[child_id].key_id]
                try:
                    if not kv.words:
                        added_key_values.add(kv.id)
                        continue
                    i = table_cells[cell_id]._children.index(kv.words[0])
                    table_cells[cell_id]._children.insert(i, kv)
                    for w in kv.words:
                        try:
                            table_cells[cell_id]._children.remove(w)
                        except ValueError:
                            continue
                    added_key_values.add(checkboxes[child_id].key_id)
                except ValueError:
                    # Word is not in the table cells words
                    continue
            # This is just a checkbox
            else:
                table_cells[cell_id]._children.append(checkboxes[child_id])

        # update metadata
        meta_info = cell.get("EntityTypes", []) or []
        merged_info = [MERGED_CELL] if cell_id in merged_child_ids else []
        table_cells[cell_id]._update_response_metadata(meta_info + merged_info)

    # This is problematic because a KV will usually not be within a single cell, but instead
    # cover multiple cells (usually one for key, one for value)
    for kv_id, kv in key_values.items():
        if kv_id in added_key_values:
            continue
        for table in page_tables:
            table = tables[table["Id"]]
            # If the kv is in the table bbox, we just drop it entirely
            if all([w in table_words for w in kv.words]):
                added_key_values.add(kv_id)

    # optimize code
    for merge_id, child_cells in merged_child_map.items():
        for child_id in child_cells:
            if child_id in table_cells.keys():
                table_cells[child_id].parent_cell_id = merge_id
                # FIXME: This should be gated
                table_cells[child_id].siblings = [
                    table_cells[cid] for cid in child_cells if cid in table_cells
                ]  # CHECK IF IDS ARE BETTER THAN INSTANCES

    # Create table title (if exists)
    for table in page_tables:
        children = _get_relationship_ids(table, relationship="TABLE_TITLE")
        for child_id in children:
            if child_id not in id_json_map:
                continue
            tables[table["Id"]].title = TableTitle(
                entity_id=child_id,
                bbox=BoundingBox.from_normalized_dict(
                    id_json_map[child_id]["Geometry"]["BoundingBox"],
                    spatial_object=page,
                ),
            )
            children = _get_relationship_ids(
                id_json_map[child_id], relationship="CHILD"
            )
            tables[table["Id"]].title.words = _create_word_objects(
                # FIXME: This should be gated
                [child_id for child_id in children if (child_id in id_entity_map and id_entity_map[child_id] == WORD)],
                id_json_map,
                existing_words,
                page,
            )

    # Create table footer (if exists)
    for table in page_tables:
        children = _get_relationship_ids(table, relationship="TABLE_FOOTER")
        for child_id in children:
            if child_id not in id_json_map:
                continue
            tables[table["Id"]].footers.append(
                TableFooter(
                    entity_id=child_id,
                    bbox=BoundingBox.from_normalized_dict(
                        id_json_map[child_id]["Geometry"]["BoundingBox"],
                        spatial_object=page,
                    ),
                )
            )
            children = _get_relationship_ids(
                id_json_map[child_id], relationship="CHILD"
            )
            tables[table["Id"]].footers[-1].words = _create_word_objects(
                [child_id for child_id in children if child_id in id_entity_map and id_entity_map[child_id] == WORD],
                id_json_map,
                existing_words,
                page,
            )

    # Associate Children with Tables
    for table in page_tables:
        children = _get_relationship_ids(table, relationship="CHILD")
        children_cells = []
        for child_id in children:
            # FIXME: This should be gated
            if child_id not in table_cells:
                continue
            children_cells.append(table_cells[child_id])
            if table_cells[child_id].is_title and tables[table["Id"]].title is not None:
                tables[table["Id"]].title.is_floating = False
        # FIXME: This will be slow and there should be a better way to do it.
        words = set()
        for child_id in children:
            if child_id not in table_cells:
                continue
            for w in table_cells[child_id].words:
                words.add(w.id)
        for footer in tables[table["Id"]].footers:
            for w in footer.words:
                if w.id in words:
                    footer.is_floating = False
                    break

        tables[table["Id"]].add_cells(children_cells)
        tables[table["Id"]].add_children(children_cells)

    # Assign tables to layout elements
    table_added = set()
    for layout in sorted(page.leaf_layouts, key=lambda x: x.bbox.y):
        if layout.layout_type == LAYOUT_TABLE:
            for table in sorted(list(tables.values()), key=lambda x: x.bbox.y):
                if layout.bbox.get_intersection(table.bbox).area > THRESHOLD*table.bbox.area and table not in table_added:
                    for w in table.words:
                        layout.remove(w)
                    layout.children.append(table)
                    layout.bbox = BoundingBox.enclosing_bbox(layout.children)
                    table_added.add(table)

    tables_layout = []
    for table in tables.values():
        if table not in table_added:
            table_added.add(table)
            layout = Layout(
                entity_id=table.id,
                bbox=table.bbox,
                label=LAYOUT_TABLE,
                reading_order=-1,
            )
            layout.children.append(table)
            layout.page = page.page_num
            layout.page_id = page.id
            tables_layout.append(layout)

    layouts_to_remove = []
    for layout in page.leaf_layouts:
        layouts_that_intersect = []
        for table_layout in tables_layout:
            intersection = layout.bbox.get_intersection(table_layout.bbox).area
            if intersection:
                layouts_that_intersect.append(table_layout)
        words_in_sub_layouts = set()
        for i, intersect_layout in enumerate(
            sorted(layouts_that_intersect, key=lambda l: (l.bbox.y, l.bbox.x))
        ):
            intersect_layout.reading_order = (
                (layout.reading_order + (i + 1) * 0.1)
                if intersect_layout.reading_order == -1
                else min(
                    intersect_layout.reading_order, layout.reading_order + (i + 1) * 0.1
                )
            )
            for w in intersect_layout.children[0].words:
                words_in_sub_layouts.add(w)
        if words_in_sub_layouts:
            for word in words_in_sub_layouts:
                layout.remove(word)
            if layout._children:
                layout.bbox = BoundingBox.enclosing_bbox(layout._children)
            else:
                layouts_to_remove.append(layout)

    for layout in layouts_to_remove:
        page.leaf_layouts.remove(layout)

    for layout in tables_layout:
        page.leaf_layouts.append(layout)

    tables = list(tables.values())
    for table in tables:
        table.page = page.page_num
        table.page_id = page.id
    return tables, table_words, added_key_values


def parse_document_api_response(response: dict) -> Document:
    """
    Parses Textract JSON response and converts them into Document object containing Page objects.
    A valid Page object must contain at least a unique name and physical dimensions.

    :param response: JSON response data in a format readable by the ResponseParser
    :type response: dict

    :return: Document object containing the hierarchy of DocumentEntity descendants.
    :rtype: Document
    """
    document = _create_document_object(response)

    id_entity_map, id_json_map, entity_id_map, existing_words = (
        {},
        {},
        defaultdict(list),
        {},
    )
    # Create de entity id map for faster lookup
    for block in response["Blocks"]:
        id_entity_map[block["Id"]] = block["BlockType"]
        id_json_map[block["Id"]] = block
        if block["BlockType"].startswith("LAYOUT"):
            entity_id_map["LAYOUT"].append(block["Id"])
        else:
            entity_id_map[block["BlockType"]].append(block["Id"])

    # Create the empty pages
    pages, page_elements = _create_page_objects(response)
    assert len(pages) == response["DocumentMetadata"]["Pages"]

    # Fill the page with the detected entities
    for page_json in page_elements:
        page = pages[page_json["Id"]]

        # Creating lines
        lines, line_words = _create_line_objects(
            entity_id_map[LINE], id_json_map, existing_words, page
        )
        page.lines = deepcopy(lines)

        line_by_id = {l.id: l for l in lines}

        # Creating layouts
        container_layouts, leaf_layouts = _create_layout_objects(
            entity_id_map[LAYOUT],
            id_json_map,
            entity_id_map,
            line_by_id,
            page,
        )

        # If no layouts were created, we create fake layouts
        if not container_layouts and not leaf_layouts:
            # We are in a scenario where the LAYOUT API was not called. We will fake wrap
            # all the lines to get a good linearized output regardless.
            for i, line in enumerate(lines):
                layout = Layout(
                    entity_id=line.id,
                    bbox=line.bbox,
                    label=LAYOUT_ENTITY,
                    reading_order=i,
                )
                layout._children = [line]
                layout.page = page.page_num
                layout.page_id = page.id
                leaf_layouts.append(layout)

        page._container_layouts.extend(container_layouts)
        page._leaf_layouts.extend(leaf_layouts)

        # Create key value objects
        key_values, kv_words, selection_elements = _create_keyvalue_objects(
            entity_id_map[KEY_VALUE_SET],
            id_json_map,
            id_entity_map,
            entity_id_map,
            existing_words,
            page,
        )
        kvs = [kv for kv in key_values if not kv.contains_checkbox]
        checkboxes = [kv for kv in key_values if kv.contains_checkbox]

        # For backward compatibility reason we split kvs and checkboxes under two attributes
        page.key_values = kvs
        page.checkboxes = checkboxes

        for checkbox in checkboxes:
            id_entity_map[checkbox.id] = SELECTION_ELEMENT

        # Create the table objects
        tables, table_words, kv_added = _create_table_objects(
            entity_id_map[TABLE],
            id_json_map,
            id_entity_map,
            entity_id_map,
            existing_words,
            # We pass the KeyValue objects because there will be overlap and we need to make
            # sure that the layout children are unique.
            {kv.id:kv for kv in key_values},
            selection_elements,
            page,
        )
        page.tables = tables

        # Using the kv_added returned by _create_table_objects, we try to match the remaining KVs
        # to existing layout elements.
        for layout in sorted(page.leaf_layouts, key=lambda x: x.bbox.y):
            if layout.layout_type == LAYOUT_ENTITY:
                continue
            for kv in sorted(key_values, key=lambda x: x.bbox.y):
                if (
                    layout.bbox.get_intersection(kv.bbox).area > THRESHOLD * kv.bbox.area
                    and kv.id not in kv_added
                ):
                    # Ignore if the KV is already overlapping with a table
                    if any([w.cell_id for w in kv.words]):
                        kv_added.add(kv.id)
                        continue
                    # Removing the duplicate words
                    for w in kv.words:
                        layout.remove(w)
                    # Adding the KV to the layout children (order is not relevant)
                    layout.children.append(kv)
                    kv_added.add(kv.id)
                    key_values.remove(kv)

        
        page.leaf_layouts = [l for l in page.leaf_layouts if l.children or l.layout_type == LAYOUT_FIGURE]

        # We create layout elements for the KeyValues that did not match to a layout element in the
        # previous step
        kv_layouts = []
        for kv in key_values:
            if kv.id not in kv_added:
                kv_added.add(kv.id)
                layout = Layout(
                    entity_id=kv.id,
                    bbox=kv.bbox,
                    label=LAYOUT_KEY_VALUE,
                    reading_order=-1,
                )
                layout.children.append(kv)
                layout.page = page.page_num
                layout.page_id = page.id
                kv_layouts.append(layout)

        # We update the existing layout elements to avoid overlap, this should only happen to
        # a few KV layouts as the previous step will have caught most overlap.
        layouts_to_remove = []
        kv_layouts_to_ignore = []
        layouts_that_intersect = defaultdict(list)
        for layout in page.leaf_layouts:
            for kv_layout in kv_layouts:
                intersection = layout.bbox.get_intersection(kv_layout.bbox).area
                if intersection:
                    layouts_that_intersect[layout].append(kv_layout)
        for layout, intersections in layouts_that_intersect.items():
            words_in_sub_layouts = set()
            for i, intersect_layout in enumerate(
                sorted(intersections, key=lambda l: (l.bbox.y, l.bbox.x))
            ):
                # If a new KV layout intersected with more than one layout, we ignore it
                if sum([intersect_layout in intsects for intsects in layouts_that_intersect.values()]) > 1:
                    kv_layouts_to_ignore.append(intersect_layout)
                    continue
                # We assign a slightly higher reading order to the intersected layout
                intersect_layout.reading_order = (
                    (layout.reading_order + (i + 1) * 0.1)
                    if intersect_layout.reading_order == -1
                    else min(
                        intersect_layout.reading_order, layout.reading_order + (i + 1) * 0.1
                    )
                )
                # We take only the first child as the intersected layout will only have the KV as
                # its child. 
                for w in intersect_layout.children[0].words:
                    words_in_sub_layouts.add(w)
            for word in words_in_sub_layouts:
                layout.remove(word)
            if not layout.children and layout.layout_type != LAYOUT_FIGURE:
                layouts_to_remove.append(layout)

        # Clean up layouts that became empty due to the previous step. 
        for layout in layouts_to_remove:
            page.leaf_layouts.remove(layout)

        # Add the new KV layouts to the page
        for layout in kv_layouts:
            if layout not in kv_layouts_to_ignore:
                page.leaf_layouts.append(layout)

        # Set the page word, create lines for orphaned words
        all_words = table_words + kv_words + line_words
        for word in all_words:
            if word.line is None:
                line = Line(
                    str(uuid.uuid4()),
                    word.bbox,
                    words=[word],
                    confidence=word.confidence,
                )
                line.page = page.page_num
                line.page_id = page.id
                word.line = line
                page.lines.append(line)
        all_words = {word.id: word for word in all_words}

        page.words = list(all_words.values())

        # Create query objects
        queries = _create_query_objects(
            entity_id_map[QUERY], id_json_map, entity_id_map, page
        )
        page.queries = queries

        # Create signature objects
        signatures = _create_signature_objects(
            entity_id_map[SIGNATURE], id_json_map, entity_id_map, page
        )
        page.signatures = signatures

        # Final clean up of the layout objects
        word_set = set()
        for layout in sorted(page.layouts, key=lambda l: l.reading_order):
            layout.visit(word_set)
            if not layout.children and layout.layout_type != LAYOUT_FIGURE:
                try:
                    page.leaf_layouts.remove(layout)
                except:
                    page.container_layouts.remove(layout)

    document.pages = sorted(list(pages.values()), key=lambda x: x.page_num)
    document.response = response
    return document

def parse_analyze_id_response(response):
    id_documents = []
    response["Blocks"] = []
    for doc in response["IdentityDocuments"]:
        fields = {}
        for field in doc["IdentityDocumentFields"]:
            fields[field["Type"]["Text"]] = {
                "key": field["Type"]["Text"],
                "value": field["ValueDetection"]["Text"],
                "confidence": field["ValueDetection"]["Confidence"],
            }
        id_documents.append(IdentityDocument(fields))
        id_documents[-1].raw_object = doc
        response["Blocks"].extend(doc.get("Blocks", []))
    # FIXME: Quick fix, we need something more robust
    document = parse_document_api_response(response)
    del response["Blocks"]
    document.identity_documents = id_documents
    document.response = response
    return document


def create_expense_from_field(field: Dict, page: Page) -> ExpenseField:
    if "Type" in field:
        type_expense = ExpenseType(
            field["Type"]["Text"], field["Type"]["Confidence"], field["Type"]
        )
    else:
        type_expense = None
    if "ValueDetection" in field:
        value_expense = Expense(
            bbox=(
                None
                if not "Geometry" in field["ValueDetection"]
                else BoundingBox.from_normalized_dict(
                    field["ValueDetection"]["Geometry"]["BoundingBox"],
                    spatial_object=page,
                )
            ),
            text=field["ValueDetection"]["Text"],
            confidence=field["ValueDetection"]["Confidence"],
            page=page.page_num,
        )
        value_expense.raw_object = field["ValueDetection"]
    else:
        value_expense = None
    if "LabelDetection" in field:
        label_expense = Expense(
            bbox=BoundingBox.from_normalized_dict(
                field["LabelDetection"]["Geometry"]["BoundingBox"], spatial_object=page
            ),
            text=field["LabelDetection"]["Text"],
            confidence=field["LabelDetection"]["Confidence"],
            page=page.page_num,
        )
        label_expense.raw_object = field["LabelDetection"]
    else:
        label_expense = None
    group_properties = []
    if "GroupProperties" in field:
        for group_property in field["GroupProperties"]:
            group_properties.append(
                ExpenseGroupProperty(
                    id=group_property["Id"], types=group_property["Types"]
                )
            )
    if "Currency" in field:
        currency = field["Currency"]["Code"]
    else:
        currency = None
    return ExpenseField(
        type_expense,
        value_expense,
        group_properties=group_properties,
        label=label_expense,
        currency=currency,
        page=page.page_num,
    )


def parser_analyze_expense_response(response):
    response["Blocks"] = [
        b for doc in response["ExpenseDocuments"] for b in doc.get("Blocks", [])
    ]
    document = parse_document_api_response(response)
    for doc in response["ExpenseDocuments"]:
        page_number = None
        if len(doc["SummaryFields"]):
            page_number = doc["SummaryFields"][0].get("PageNumber")
        elif len(doc["LineItemGroups"]):
            # A group must have at least one LineItem, and a LI must have at least one field:
            first_field = doc["LineItemGroups"][0]["LineItems"][0]["LineItemExpenseFields"][0]
            page_number = first_field.get("PageNumber")
        if page_number is None:
            logging.warning(
                "Skipping parsing ExpenseDocument %s as its page number could not be determined"
                % (doc["ExpenseIndex"],)
            )
            continue

        page = document.pages[page_number - 1]
        summary_fields = []
        for summary_field in doc["SummaryFields"]:
            summary_fields.append(create_expense_from_field(summary_field, page))
            summary_fields[-1].raw_object = summary_field

        line_items_groups = []
        for line_items_group in doc["LineItemGroups"]:
            line_item_rows = []
            for i, line_item in enumerate(line_items_group["LineItems"]):
                row_expenses = []
                for line_item_field in line_item["LineItemExpenseFields"]:
                    row_expenses.append(
                        create_expense_from_field(line_item_field, page)
                    )
                    row_expenses[-1].raw_object = line_item_field
                line_item_rows.append(
                    LineItemRow(
                        index=i,
                        line_item_expense_fields=row_expenses,
                        page=page.page_num,
                    )
                )
            if not line_item_rows:
                continue
            line_items_groups.append(
                LineItemGroup(
                    index=line_items_group["LineItemGroupIndex"],
                    line_item_rows=line_item_rows,
                    page=page.page_num,
                )
            )

        bbox = BoundingBox.enclosing_bbox(
            bboxes=[s.bbox for s in summary_fields]
            + [g.bbox for g in line_items_groups],
            spatial_object=page,
        )
        expense_document = ExpenseDocument(
            summary_fields=summary_fields,
            line_items_groups=line_items_groups,
            bounding_box=bbox,
            page=page.page_num,
        )
        expense_document.raw_object = doc
        document.pages[page_number - 1].expense_documents.append(
            expense_document
        )
    document.response = response
    return document


def parse(response: dict) -> Document:
    """
    Ingests response data and API Call Mode and calls the appropriate function for it.
    Presently supports only SYNC and ASYNC API calls. Will be extended to Analyze ID and Expense in the future.

    :param response: JSON response data in a format readable by the ResponseParser.
    :type response: dict

    :return: Document object returned after making respective parse function calls.
    :rtype: Document
    """
    if "IdentityDocuments" in response:
        return parse_analyze_id_response(response)
    if "ExpenseDocuments" in response:
        return parser_analyze_expense_response(response)
    else:
        return parse_document_api_response(response)
